import functools
import logging
import sys
from copy import deepcopy
from enum import Enum
from json import JSONDecodeError
from pathlib import Path
from threading import Event, Lock
from time import sleep
from typing import Any, ClassVar, NewType, Optional, Tuple

import paho.mqtt.client as mqtt
from paho.mqtt.client import Client as MQTT
from paho.mqtt.client import MQTTMessage
from pydantic import BaseModel, ValidationError

from .mqtt import MQTTClient

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

ASTORIA_VERSION = '0.11.2'

LOGGER = logging.getLogger(__name__)
CONFIG_SEARCH_PATHS = [
    Path("astoria.toml"),
    Path("/etc/astoria.toml"),
]


# Astoria config

class MQTTBrokerInfo(BaseModel):
    """MQTT Broker Information."""
    class Config:
        extra = 'forbid'

    host: str
    port: int
    enable_tls: bool = False
    topic_prefix: str = "astoria"
    force_protocol_version_3_1: bool = False


class AstoriaConfig(BaseModel):
    """Config schema for Astoria."""

    mqtt: MQTTBrokerInfo

    @classmethod
    def _get_config_path(cls, config_path: Optional[str] = None) -> Path:
        """Check for a config file or search the filesystem for one."""
        if config_path is None:
            for path in CONFIG_SEARCH_PATHS:
                if path.is_file():
                    return path
        else:
            path = Path(config_path)
            if path.is_file():
                return path
        raise FileNotFoundError("Unable to find config file.")

    @classmethod
    def load(cls, config_str: Optional[str] = None) -> "AstoriaConfig":
        """Load the config."""
        config_path = cls._get_config_path(config_str)
        with config_path.open("rb") as fh:
            return cls.parse_obj(tomllib.load(fh))


# Metadata

class RobotMode(Enum):
    """Running Status of the manager daemon."""

    COMP = "COMP"
    DEV = "DEV"


class Metadata(BaseModel):
    """
    Astoria Metadata.

    As the metadata is passed into a templating engine for initial log lines,
    please do not add nested fields to this schema.
    """
    class Config:
        # Revalidate the model when fields are changed
        validate_assignment = True

    # From Meta USB
    arena: str = "A"
    zone: int = 0
    mode: RobotMode = RobotMode.DEV
    marker_offset: int = 0
    game_timeout: Optional[int] = None
    wifi_enabled: bool = True

    # From robot settings file
    wifi_ssid: Optional[str] = None
    wifi_psk: Optional[str] = None
    wifi_region: Optional[str] = None


# Process manager fields

class CodeStatus(str, Enum):
    """Status of the running code."""

    STARTING = "code_starting"
    RUNNING = "code_running"
    KILLED = "code_killed"
    FINISHED = "code_finished"
    CRASHED = "code_crashed"


DiskUUID = NewType("DiskUUID", str)


class DiskType(Enum):
    """Type of disk."""

    USERCODE = "USERCODE"
    METADATA = "METADATA"
    NOACTION = "NOACTION"


class DiskInfo(BaseModel):
    """Information about a mounted disk."""

    uuid: DiskUUID
    mount_path: Path
    disk_type: DiskType


# Manager messages

class ManagerMessage(BaseModel):
    """Common data that all manager messages output."""

    class Status(Enum):
        """Running Status of the manager daemon."""

        STOPPED = "STOPPED"
        RUNNING = "RUNNING"

    status: Status
    astoria_version: str = ASTORIA_VERSION


class ProcessManagerMessage(ManagerMessage):
    """
    Status message for Process Manager.

    Published to astoria/astprocd
    """

    code_status: Optional[CodeStatus]
    disk_info: Optional[DiskInfo]
    pid: Optional[int]


class MetadataManagerMessage(ManagerMessage):
    """
    Status message for Metadata Manager.

    Published to /astoria/astmetad
    """

    metadata: Metadata


# Remote start event
@functools.total_ordering
class StartButtonBroadcastEvent(BaseModel):
    """
    Schema for a remote start event.

    Trigger the robot code if it is waiting for the start button to be
    pressed. Does not affect or interact with the physical button as that
    is handled by the usercode driver.
    """

    name: ClassVar[str] = "start_button"

    event_name: str
    sender_name: str
    priority: int = 0

    def __gt__(self, other: "StartButtonBroadcastEvent") -> bool:
        return self.priority > other.priority


# Module Logic

class AstoriaInterface:
    def __init__(self, mqtt_client: MQTTClient) -> None:
        self._mqtt = mqtt_client
        self._metadata = Metadata()
        self._mount_path = Path("/dev/null")
        self._astoria_lock = Lock()
        self._start_pressed = Event()

        self._mqtt.subscribe("broadcast/start_button", self._process_remote_start)
        self._mqtt.subscribe("astmetad", self._process_metadata_update)
        self._mqtt.subscribe("astprocd", self._handle_astprocd_message)
        # Wait a short time for the mount path to be updated
        sleep(0.05)

    def _process_remote_start(
        self, client: MQTT, userdata: Any, msg: MQTTMessage,
    ) -> None:
        """Receive a remote start event."""
        try:
            event: StartButtonBroadcastEvent = StartButtonBroadcastEvent.parse_raw(msg.payload)
            LOGGER.debug(f"Received a {event.event_name} event from {event.sender_name}")
            self._start_pressed.set()
        except (JSONDecodeError, ValidationError) as e:
            LOGGER.debug(f"Unable to parse start button message: {e}")

    def _process_metadata_update(
        self, client: MQTT, userdata: Any, msg: MQTTMessage,
    ) -> None:
        """Update the metadata."""
        try:
            MetadataManagerMessage.update_forward_refs()
            message = MetadataManagerMessage.parse_raw(msg.payload)
            if message.status == MetadataManagerMessage.Status.RUNNING:
                LOGGER.debug("Received metadata")
                with self._astoria_lock:
                    self._metadata = message.metadata
            else:
                LOGGER.debug("Cannot get metadata, astmetad is not running")
        except (JSONDecodeError, ValidationError) as e:
            LOGGER.debug(f"Unable to parse metadata message: {e}")

    def _handle_astprocd_message(
        self, client: MQTT, userdata: Any, msg: MQTTMessage,
    ) -> None:
        """Update the mount path."""
        try:
            message = ProcessManagerMessage.parse_raw(msg.payload)
            if message.status == ProcessManagerMessage.Status.RUNNING:
                LOGGER.debug("Received process info")
                if message.disk_info:
                    with self._astoria_lock:
                        self._mount_path = message.disk_info.mount_path
            else:
                LOGGER.debug("Cannot get process info, astprocd is not running")
        except (JSONDecodeError, ValidationError) as e:
            LOGGER.debug(f"Unable to parse process message: {e}")

    def fetch_current_metadata(self) -> Metadata:
        """Fetch the current metadata."""
        with self._astoria_lock:
            return deepcopy(self._metadata)

    def fetch_mount_path(self) -> Path:
        """Fetch the current mount path."""
        with self._astoria_lock:
            return self._mount_path

    def get_start_button_pressed(self) -> bool:
        """Get the start button pressed status."""
        pressed = self._start_pressed.is_set()
        self._start_pressed.clear()
        return pressed


def init_mqtt(config: AstoriaConfig, client_name: str = 'sr-robot3') -> 'MQTTClient':
    """
    Helper method to create an MQTT client and connect.

    Uses values from astoria's config.
    """
    client = MQTTClient.establish(
        host=config.mqtt.host,
        port=config.mqtt.port,
        client_name=client_name,
        mqtt_version=(
            mqtt.MQTTProtocolVersion.MQTTv311
            if config.mqtt.force_protocol_version_3_1 else
            mqtt.MQTTProtocolVersion.MQTTv5
        ),
        topic_prefix=config.mqtt.topic_prefix,
    )
    return client


def init_astoria_mqtt() -> Tuple[MQTTClient, AstoriaInterface]:
    """Initialise the MQTT client & Astoria integrations."""
    try:
        astoria_config = AstoriaConfig.load()
    except FileNotFoundError:
        LOGGER.warning("Unable to find astoria.toml, using default values")
        astoria_config = AstoriaConfig(mqtt=MQTTBrokerInfo(host="localhost", port=1883))

    mqtt_client = init_mqtt(astoria_config)
    astoria = AstoriaInterface(mqtt_client)

    return mqtt_client, astoria



"""KCH Driver."""
from __future__ import annotations

import atexit
import logging
import warnings
from enum import IntEnum, unique
from typing import Optional

from .exceptions import BoardDisconnectionError, IncorrectBoardError
from .serial_wrapper import SerialWrapper
from .utils import IN_SIMULATOR, Board, BoardIdentity, get_simulator_boards

try:
    import RPi.GPIO as GPIO  # isort: ignore
    HAS_HAT = True if not IN_SIMULATOR else False
except ImportError:
    HAS_HAT = False


logger = logging.getLogger(__name__)

# Only used in the simulator
BAUDRATE = 115200


@unique
class RobotLEDs(IntEnum):
    """Mapping of LEDs to GPIO Pins."""

    START = 9

    USER_A_RED = 24
    USER_A_GREEN = 10
    USER_A_BLUE = 25
    USER_B_RED = 27
    USER_B_GREEN = 23
    USER_B_BLUE = 22
    USER_C_RED = 4
    USER_C_GREEN = 18
    USER_C_BLUE = 17

    @classmethod
    def all_leds(cls) -> list[int]:
        """Get all LEDs."""
        return [c.value for c in cls]

    @classmethod
    def user_leds(cls) -> list[tuple[int, int, int]]:
        """Get the user programmable LEDs."""
        return [
            (cls.USER_A_RED, cls.USER_A_GREEN, cls.USER_A_BLUE),
            (cls.USER_B_RED, cls.USER_B_GREEN, cls.USER_B_BLUE),
            (cls.USER_C_RED, cls.USER_C_GREEN, cls.USER_C_BLUE),
        ]


@unique
class UserLED(IntEnum):
    """User Programmable LEDs."""

    A = 0
    B = 1
    C = 2


class Colour():
    """User LED colours."""

    OFF = (False, False, False)
    RED = (True, False, False)
    YELLOW = (True, True, False)
    GREEN = (False, True, False)
    CYAN = (False, True, True)
    BLUE = (False, False, True)
    MAGENTA = (True, False, True)
    WHITE = (True, True, True)


class KCH:
    """KCH Board."""
    __slots__ = ('_leds', '_pwm')

    def __init__(self) -> None:
        self._leds: tuple[LED, ...]

        if HAS_HAT:
            GPIO.setmode(GPIO.BCM)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                # If this is not the first time the code is run this init will
                # cause a warning as the gpio are already initialized, we can
                # suppress this as we know the reason behind the warning
                GPIO.setup(RobotLEDs.all_leds(), GPIO.OUT, initial=GPIO.LOW)
            self._pwm: Optional[GPIO.PWM] = None

            # We are not running cleanup so the LED state persists after the code completes,
            # this will cause a warning with `GPIO.setup()` which we can suppress
            # atexit.register(GPIO.cleanup)

            # Cleanup just the start LED to turn it off when the code exits
            # Mypy isn't aware of the version of atexit.register(func, *args)
            atexit.register(GPIO.cleanup, RobotLEDs.START)  # type: ignore[call-arg]

            self._leds = tuple(
                PhysicalLED(rgb_led) for rgb_led in RobotLEDs.user_leds()
            )
        elif IN_SIMULATOR:
            led_server = LedServer.initialise()
            if led_server is not None:
                self._leds = tuple(
                    SimulationLED(v, led_server)
                    for v in range(len(RobotLEDs.user_leds()))
                )
            else:
                self._leds = tuple(
                    LED() for _ in RobotLEDs.user_leds()
                )
        else:
            self._leds = tuple(
                LED() for _ in RobotLEDs.user_leds()
            )

    @property
    def _start(self) -> bool:
        """Get the state of the start LED."""
        return GPIO.input(RobotLEDs.START) if HAS_HAT else False

    @_start.setter
    def _start(self, value: bool) -> None:
        """Set the state of the start LED."""
        if HAS_HAT:
            if self._pwm:
                # stop any flashing the LED is doing
                self._pwm.stop()
                self._pwm = None
            GPIO.output(RobotLEDs.START, GPIO.HIGH if value else GPIO.LOW)

    def _flash_start(self) -> None:
        """Set the start LED flashing at 1Hz."""
        if HAS_HAT:
            self._pwm = GPIO.PWM(RobotLEDs.START, 1)
            self._pwm.start(50)

    @property
    def leds(self) -> tuple['LED', ...]:
        """User programmable LEDs."""
        return self._leds


class LED:
    """
    User programmable LED.

    This is a dummy class to handle the case where this is run on neither the
    Raspberry Pi nor the simulator.
    As such, this class does nothing.
    """
    __slots__ = ('_led',)

    @property
    def r(self) -> bool:
        """Get the state of the Red LED segment."""
        return False

    @r.setter
    def r(self, value: bool) -> None:
        """Set the state of the Red LED segment."""

    @property
    def g(self) -> bool:
        """Get the state of the Green LED segment."""
        return False

    @g.setter
    def g(self, value: bool) -> None:
        """Set the state of the Green LED segment."""

    @property
    def b(self) -> bool:
        """Get the state of the Blue LED segment."""
        return False

    @b.setter
    def b(self, value: bool) -> None:
        """Set the state of the Blue LED segment."""

    @property
    def colour(self) -> tuple[bool, bool, bool]:
        """Get the colour of the user LED."""
        return False, False, False

    @colour.setter
    def colour(self, value: tuple[bool, bool, bool]) -> None:
        """Set the colour of the user LED."""
        if not isinstance(value, (tuple, list)) or len(value) != 3:
            raise ValueError("The LED requires 3 values for it's colour")


class PhysicalLED(LED):
    """
    User programmable LED.

    Used when running on the Raspberry Pi to control the actual LEDs.
    """
    __slots__ = ('_led',)

    def __init__(self, led: tuple[int, int, int]):
        self._led = led

    @property
    def r(self) -> bool:
        """Get the state of the Red LED segment."""
        return GPIO.input(self._led[0])

    @r.setter
    def r(self, value: bool) -> None:
        """Set the state of the Red LED segment."""
        GPIO.output(self._led[0], GPIO.HIGH if value else GPIO.LOW)

    @property
    def g(self) -> bool:
        """Get the state of the Green LED segment."""
        return GPIO.input(self._led[1])

    @g.setter
    def g(self, value: bool) -> None:
        """Set the state of the Green LED segment."""
        GPIO.output(self._led[1], GPIO.HIGH if value else GPIO.LOW)

    @property
    def b(self) -> bool:
        """Get the state of the Blue LED segment."""
        return GPIO.input(self._led[2])

    @b.setter
    def b(self, value: bool) -> None:
        """Set the state of the Blue LED segment."""
        GPIO.output(self._led[2], GPIO.HIGH if value else GPIO.LOW)

    @property
    def colour(self) -> tuple[bool, bool, bool]:
        """Get the colour of the user LED."""
        return (
            GPIO.input(self._led[0]),
            GPIO.input(self._led[1]),
            GPIO.input(self._led[2]),
        )

    @colour.setter
    def colour(self, value: tuple[bool, bool, bool]) -> None:
        """Set the colour of the user LED."""
        if not isinstance(value, (tuple, list)) or len(value) != 3:
            raise ValueError("The LED requires 3 values for it's colour")

        GPIO.output(
            self._led,
            tuple(
                GPIO.HIGH if v else GPIO.LOW for v in value
            ),
        )


class LedServer(Board):
    """
    LED control over a socket.

    Used when running in the simulator to control the simulated LEDs.
    """

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'KCHv1B'.
        """
        return 'KCHv1B'

    def __init__(
        self,
        serial_port: str,
        initial_identity: BoardIdentity | None = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()
        self._serial = SerialWrapper(
            serial_port,
            BAUDRATE,
            identity=initial_identity,
        )

        self._identity = self.identify()
        if self._identity.board_type != self.get_board_type():
            raise IncorrectBoardError(self._identity.board_type, self.get_board_type())
        self._serial.set_identity(self._identity)

        # Reset the board to a known state
        self._serial.write('*RESET')

    @classmethod
    def initialise(cls) -> 'LedServer' | None:
        """Initialise the LED server using simulator discovery."""
        # The filter here is the name of the emulated board in the simulator
        boards = get_simulator_boards('LedBoard')

        if not boards:
            return None

        board_info = boards[0]

        # Create board identity from the info given
        initial_identity = BoardIdentity(
            manufacturer='sbot_simulator',
            board_type=board_info.type_str,
            asset_tag=board_info.serial_number,
        )

        try:
            board = cls(board_info.url, initial_identity)
        except BoardDisconnectionError:
            logger.warning(
                f"Simulator specified LED board at port {board_info.url!r}, "
                "could not be identified. Ignoring this device")
            return None
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None

        return board

    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        :return: The identity of the board.
        """
        response = self._serial.query('*IDN?')
        return BoardIdentity(*response.split(':'))

    def set_leds(self, led_num: int, value: tuple[bool, bool, bool]) -> None:
        """Set the colour of the LED."""
        self._serial.write(f'LED:{led_num}:SET:{value[0]:d}:{value[1]:d}:{value[2]:d}')

    def get_leds(self, led_num: int) -> tuple[bool, bool, bool]:
        """Get the colour of the LED."""
        response = self._serial.query(f'LED:{led_num}:GET?')
        red, green, blue = response.split(':')
        return bool(int(red)), bool(int(green)), bool(int(blue))


class SimulationLED(LED):
    """
    User programmable LED.

    Used when running in the simulator to control the simulated LEDs.
    """
    __slots__ = ('_led_num', '_server')

    def __init__(self, led_num: int, server: LedServer) -> None:
        self._led_num = led_num
        self._server = server

    @property
    def r(self) -> bool:
        """Get the state of the Red LED segment."""
        return self._server.get_leds(self._led_num)[0]

    @r.setter
    def r(self, value: bool) -> None:
        """Set the state of the Red LED segment."""
        # Fetch the current state of the LED so we can update only the red value
        current = self._server.get_leds(self._led_num)
        self._server.set_leds(self._led_num, (value, current[1], current[2]))

    @property
    def g(self) -> bool:
        """Get the state of the Green LED segment."""
        return self._server.get_leds(self._led_num)[1]

    @g.setter
    def g(self, value: bool) -> None:
        """Set the state of the Green LED segment."""
        # Fetch the current state of the LED so we can update only the green value
        current = self._server.get_leds(self._led_num)
        self._server.set_leds(self._led_num, (current[0], value, current[2]))

    @property
    def b(self) -> bool:
        """Get the state of the Blue LED segment."""
        return self._server.get_leds(self._led_num)[2]

    @b.setter
    def b(self, value: bool) -> None:
        """Set the state of the Blue LED segment."""
        # Fetch the current state of the LED so we can update only the blue value
        current = self._server.get_leds(self._led_num)
        self._server.set_leds(self._led_num, (current[0], current[1], value))

    @property
    def colour(self) -> tuple[bool, bool, bool]:
        """Get the colour of the user LED."""
        return self._server.get_leds(self._led_num)

    @colour.setter
    def colour(self, value: tuple[bool, bool, bool]) -> None:
        """Set the colour of the user LED."""
        if not isinstance(value, (tuple, list)) or len(value) != 3:
            raise ValueError("The LED requires 3 values for it's colour")

        self._server.set_leds(self._led_num, value)



from __future__ import annotations

import atexit
import json
import logging
import time
from threading import Lock
from typing import Any, Callable, Optional, TypedDict, Union
from urllib.parse import urlparse

import paho.mqtt.client as mqtt

LOGGER = logging.getLogger(__name__)


class MQTTClient:
    def __init__(
        self,
        client_name: Optional[str] = None,
        topic_prefix: Optional[str] = None,
        mqtt_version: mqtt.MQTTProtocolVersion = mqtt.MQTTProtocolVersion.MQTTv5,
        use_tls: Union[bool, str] = False,
        username: str = '',
        password: str = '',
    ) -> None:
        self._subscriptions_lock = Lock()
        self._subscriptions: dict[
            str, Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None]
        ] = {}
        self.topic_prefix = topic_prefix
        self._client_name = client_name

        self._client = mqtt.Client(
            callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
            client_id=client_name,
            protocol=mqtt_version,
        )
        self._client.on_connect = self._on_connect

        if use_tls:
            self._client.tls_set()
            if use_tls == 'insecure':
                self._client.tls_insecure_set(True)

        if username:
            self._client.username_pw_set(username, password)

    def connect(self, host: str, port: int) -> None:
        """
        Connect to the MQTT broker and start event loop in background thread.

        Registers an atexit routine that tears down the client.
        """
        if self._client.is_connected():
            LOGGER.error("Attempting connection, but client is already connected.")
            return

        try:
            self._client.connect(host, port, keepalive=60)
        except (TimeoutError, ValueError, ConnectionRefusedError):
            LOGGER.error(f"Failed to connect to MQTT broker at {host}:{port}")
            return
        self._client.loop_start()
        atexit.unregister(self.disconnect)  # Avoid duplicate atexit handlers
        atexit.register(self.disconnect)

    @classmethod
    def establish(
        cls, host: str, port: int, **kwargs: Any,
    ) -> 'MQTTClient':
        """Create client and connect."""
        client = cls(**kwargs)
        client.connect(host, port)
        return client

    def disconnect(self) -> None:
        """Disconnect from the broker and close background event loop."""
        self._client.disconnect()
        self._client.loop_stop()
        atexit.unregister(self.disconnect)

    def subscribe(
        self,
        topic: str,
        callback: Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None],
        abs_path: bool = False,
    ) -> None:
        """Subscribe to a topic and assign a callback for messages."""
        if not abs_path and self.topic_prefix is not None:
            full_topic = f"{self.topic_prefix}/{topic}"
        else:
            full_topic = topic

        with self._subscriptions_lock:
            self._subscriptions[full_topic] = callback
        self._subscribe(full_topic, callback)

    def _subscribe(
        self,
        topic: str,
        callback: Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None],
    ) -> None:
        LOGGER.debug(f"Subscribing to {topic}")
        self._client.message_callback_add(topic, callback)
        self._client.subscribe(topic, qos=1)

    def unsubscribe(self, topic: str) -> None:
        """Unsubscribe from a topic."""
        try:
            with self._subscriptions_lock:
                del self._subscriptions[topic]
        except KeyError:
            pass
        self._client.message_callback_remove(topic)
        self._client.unsubscribe(topic)

    def publish(
        self,
        topic: str,
        payload: Union[bytes, str],
        retain: bool = False,
        *,
        abs_topic: bool = False,
    ) -> None:
        """Publish a message to the broker."""
        if not self._client.is_connected():
            LOGGER.debug(
                "Attempted to publish message, but client is not connected.",
            )
            return

        if not abs_topic and self.topic_prefix:
            topic = f"{self.topic_prefix}/{topic}"

        try:
            self._client.publish(topic, payload=payload, retain=retain, qos=1)
        except ValueError as e:
            raise ValueError(f"Cannot publish to MQTT topic: {topic}") from e

    def wrapped_publish(
        self,
        topic: str,
        payload: Union[bytes, str],
        retain: bool = False,
        *,
        abs_topic: bool = True,
    ) -> None:
        """Wrap a payload up to be decodable as JSON."""
        if isinstance(payload, bytes):
            payload = payload.decode('utf-8')

        payload_dict = {
            "timestamp": time.time(),
            "data": payload,
        }

        self.publish(
            topic,
            json.dumps(payload_dict),
            retain=retain, abs_topic=abs_topic)

    def _on_connect(
        self,
        client: mqtt.Client,
        userdata: Any,
        connect_flags: mqtt.ConnectFlags,
        reason_code: mqtt.ReasonCode,
        properties: mqtt.Properties | None = None,
    ) -> None:
        """Callback run each time the client connects to the broker."""
        if reason_code.is_failure:
            LOGGER.warning(
                f"Failed to connect to MQTT broker. Return code: {reason_code.getName()}"  # type: ignore[no-untyped-call] # noqa: E501
            )
            return

        LOGGER.debug("Connected to MQTT broker.")

        with self._subscriptions_lock:
            for topic, callback in self._subscriptions.items():
                self._subscribe(topic, callback)


class MQTTVariables(TypedDict):
    host: str
    port: int
    topic_prefix: str
    use_tls: Union[bool, str]
    username: Optional[str]
    password: Optional[str]


def unpack_mqtt_url(url: str) -> MQTTVariables:
    """
    Unpack an MQTT URL into its components to be passed to MQTTClient.

    The URL should be in the format:
    mqtt[s]://[<username>[:<password>]]@<host>[:<port>]/<topic_root>
    """
    url_parts = urlparse(url, allow_fragments=False)

    if url_parts.scheme not in ('mqtt', 'mqtts'):
        raise ValueError(f"Invalid scheme: {url_parts.scheme}")

    use_tls = (url_parts.scheme == 'mqtts')

    if url_parts.hostname is None:
        raise ValueError("No hostname given")

    return MQTTVariables(
        host=url_parts.hostname,
        port=url_parts.port or (8883 if use_tls else 1883),
        topic_prefix=url_parts.path,
        use_tls=use_tls,
        username=url_parts.username,
        password=url_parts.password,
    )



"""
Test that the arduino board can be created and used.

This test uses a mock serial wrapper to simulate the connection to the arduino board.
"""
from __future__ import annotations

import logging
from typing import NamedTuple

import pytest

from sr.robot3.arduino import AnalogPins, Arduino, GPIOPinMode
from sr.robot3.utils import BoardIdentity, singular

from .conftest import MockSerialWrapper


class MockArduino(NamedTuple):
    """A mock arduino board."""

    serial_wrapper: MockSerialWrapper
    arduino_board: Arduino


@pytest.fixture
def arduino_serial(monkeypatch) -> None:
    serial_wrapper = MockSerialWrapper([
        ("v", "SRduino:2.0"),  # Called by Arduino.__init__
    ])
    monkeypatch.setattr('sr.robot3.arduino.SerialWrapper', serial_wrapper)
    arduino_board = Arduino('test://', initial_identity=BoardIdentity(asset_tag='TEST123'))

    yield MockArduino(serial_wrapper, arduino_board)

    # Test that we made all the expected calls
    assert serial_wrapper.request_index == len(serial_wrapper.responses)


def test_arduino_board_identify(arduino_serial: MockArduino) -> None:
    """
    Test that we can create a arduino board with a mock serial wrapper.

    Uses the identify method to test that the mock serial wrapper is working.
    """
    serial_wrapper = arduino_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("v", "SSRduino:2.0"),
    ])
    arduino_board = arduino_serial.arduino_board

    # Test that the port was correctly passed to the mock serial wrapper init
    assert serial_wrapper._port == 'test://'

    # Test that the identity is correctly set from the first *IDN? response
    assert arduino_board._identity.board_type == "SRduino"
    assert arduino_board._identity.asset_tag == "TEST123"

    # Test identify returns the identity
    assert arduino_board.identify().asset_tag == "TEST123"


def test_arduino_custom_command(arduino_serial: MockArduino) -> None:
    """
    Test that we can send a custom command to the arduino board.

    This test uses the mock serial wrapper to simulate the arduino.
    """
    arduino = arduino_serial.arduino_board
    arduino_serial.serial_wrapper._add_responses([
        ("test:f", "ACK123"),
    ])

    # Test that we can send a custom command
    pin_str = arduino.map_pin_number(5)
    assert pin_str == "f"  # zero indexed from 'a'
    assert arduino.command(f"test:{pin_str}") == "ACK123"


def test_arduino_pins(arduino_serial: MockArduino) -> None:
    """
    Test the arduino pins properties and methods.

    This test uses the mock serial wrapper to simulate the arduino.
    """
    arduino = arduino_serial.arduino_board
    arduino_serial.serial_wrapper._add_responses([
        ("oc", ""),  # pin 2 is set to output
        ("pk", ""),  # pin 10 is set to input pullup
        ("io", ""),  # pin 14 is set to input
    ])

    with pytest.raises(IOError):
        arduino.pins[2].mode = 1

    # Test that we can set the mode of a pin
    arduino.pins[2].mode = GPIOPinMode.OUTPUT
    arduino.pins[10].mode = GPIOPinMode.INPUT_PULLUP
    arduino.pins[AnalogPins.A0].mode = GPIOPinMode.INPUT

    # Test that we can get the mode of a pin
    assert arduino.pins[2].mode == GPIOPinMode.OUTPUT
    assert arduino.pins[10].mode == GPIOPinMode.INPUT_PULLUP
    assert arduino.pins[AnalogPins.A0].mode == GPIOPinMode.INPUT

    # Test that we can get the digital value of a pin
    arduino_serial.serial_wrapper._add_responses([
        ("rc", "h"),  # pin 2 read
        ("rk", "l"),  # pin 10 read
        ("ro", "h"),  # pin 14 read
    ])
    assert arduino.pins[2].digital_read() is True
    assert arduino.pins[10].digital_read() is False
    assert arduino.pins[AnalogPins.A0].digital_read() is True

    # Test that we can set the digital value of a pin
    arduino_serial.serial_wrapper._add_responses([
        ("hc", "ACK"),  # pin 2 write
        ("lc", "ACK"),  # pin 2 write
    ])
    arduino.pins[2].digital_write(True)
    arduino.pins[2].digital_write(False)
    with pytest.raises(IOError, match=r"Digital write is not supported.*"):
        arduino.pins[10].digital_write(False)
    with pytest.raises(IOError, match=r"Digital write is not supported.*"):
        arduino.pins[AnalogPins.A0].digital_write(True)

    # Test that we can get the analog value of a pin
    arduino_serial.serial_wrapper._add_responses([
        ("il", ""),  # pin 11 is set to input
        ("ao", "1000"),  # pin 14 analog read
    ])
    with pytest.raises(IOError, match=r"Analog read is not supported.*"):
        arduino.pins[2].analog_read()

    arduino.pins[11].mode = GPIOPinMode.INPUT
    with pytest.raises(IOError, match=r"Pin does not support analog read"):
        arduino.pins[11].analog_read()
    # 4.888 = round((5 / 1023) * 1000, 3)
    assert arduino.pins[AnalogPins.A0].analog_read() == 4.888


def test_invalid_properties(arduino_serial: MockArduino) -> None:
    """
    Test that settng invalid properties raise an AttributeError.
    """
    arduino_board = arduino_serial.arduino_board

    with pytest.raises(AttributeError):
        arduino_board.invalid_property = 1

    with pytest.raises(AttributeError):
        arduino_board.pins.invalid_property = 1

    with pytest.raises(AttributeError):
        arduino_board.pins[0].invalid_property = 1


def test_arduino_board_discovery(monkeypatch) -> None:
    """
    Test that discovery finds arduino boards from USB serial ports.

    Test that different USB pid/vid combinations are ignored.
    """
    class ListPortInfo(NamedTuple):
        """A mock serial port info."""
        device: str
        manufacturer: str
        product: str
        serial_number: str
        vid: int
        pid: int

    def mock_comports() -> list[ListPortInfo]:
        ports = [
            ListPortInfo(
                device='test://1',
                manufacturer='Student Robotics',
                product='Arduino',
                serial_number='TEST123',
                vid=0x2341,
                pid=0x0043,
            ),
            ListPortInfo(  # A arduino board with a different pid/vid
                device='test://3',
                manufacturer='Other',
                product='Arduino',
                serial_number='OTHER',
                vid=0x1234,
                pid=0x5678,
            ),
            ListPortInfo(  # An unrelated device
                device='test://5',
                manufacturer='Student Robotics',
                product='OTHER',
                serial_number='TESTABC',
                vid=0x2341,
                pid=0x0043,
            ),
        ]
        return ports

    serial_wrapper = MockSerialWrapper([
        ("v", "SRduino:2.0"),  # USB discovered board
        ("v", "Arduino:4.3"),  # USB invalid board
        ("v", "SRcustom:3.0"),  # Manually added board
        ("v", "OTHER:4.3"),  # Manual invalid board
        ("v", "SRduino:2.0"),  # identity check
        ("v", "SRcustom:3.0"),  # identity check
    ])
    monkeypatch.setattr('sr.robot3.arduino.SerialWrapper', serial_wrapper)
    monkeypatch.setattr('sr.robot3.arduino.comports', mock_comports)

    arduino_boards = Arduino._get_supported_boards(manual_boards=['test://2', 'test://4'])
    assert len(arduino_boards) == 2
    # Manually added boards use the serial port as the asset tag
    assert {'TEST123', 'test://2'} == set(arduino_boards.keys())

    identity = arduino_boards['TEST123'].identify()
    assert identity.board_type == "SRduino"
    assert identity.sw_version == "2.0"

    identity = arduino_boards['test://2'].identify()
    assert identity.board_type == "SRcustom"
    assert identity.sw_version == "3.0"


def test_arduino_board_new_sourcebots_identity(monkeypatch, caplog) -> None:
    """
    Test that we raise an error if the arduino board returns an old style identity.
    """
    serial_wrapper = MockSerialWrapper([
        ("v", "NACK:Invalid command"),  # Called by Arduino.__init__
    ])
    monkeypatch.setattr('sr.robot3.arduino.SerialWrapper', serial_wrapper)

    with caplog.at_level(logging.WARNING):
        arduino = Arduino._get_valid_board('test://', BoardIdentity(board_type="manual"))

    assert arduino is None
    assert caplog.records[0].message == (
        "Board returned type 'NACK', expected 'SR*'. Ignoring this device"
    )


def test_arduino_board_old_sourcebots_identity(monkeypatch, caplog) -> None:
    """
    Test that we raise an error if the arduino board returns an old style identity.
    """
    serial_wrapper = MockSerialWrapper([
        ("v", "Error, unknown command: v"),  # Called by Arduino.__init__
    ])
    monkeypatch.setattr('sr.robot3.arduino.SerialWrapper', serial_wrapper)

    with caplog.at_level(logging.WARNING):
        arduino = Arduino._get_valid_board('test://', BoardIdentity(board_type="manual"))

    assert arduino is None
    assert caplog.records[0].message == (
        "Board returned type 'Error, unknown command', expected 'SR*'. Ignoring this device"
    )


@pytest.mark.hardware
def test_physical_arduino_board_discovery() -> None:
    """Test that we can discover physical arduino boards."""
    arduino_boards = Arduino._get_supported_boards()
    assert len(arduino_boards) == 1, "Did not find exactly one arduino board."
    arduino_board = singular(arduino_boards)
    identity = arduino_board.identify()
    assert identity.board_type == "Arduino", "arduino board is not the correct type."
    asset_tag = identity.asset_tag
    assert arduino_board == arduino_boards[asset_tag], \
        "Singular arduino board is not the same as the one in the list of arduino boards."



"""
Test that the power board can be created and used.

This test uses a mock serial wrapper to simulate the connection to the power board.
"""
from __future__ import annotations

import re
from typing import NamedTuple

import pytest

from sr.robot3.exceptions import IncorrectBoardError
from sr.robot3.power_board import Note, PowerBoard, PowerOutputPosition
from sr.robot3.utils import singular

from .conftest import MockAtExit, MockSerialWrapper


class MockPowerBoard(NamedTuple):
    """A mock power board."""

    serial_wrapper: MockSerialWrapper
    power_board: PowerBoard


@pytest.fixture
def powerboard_serial(monkeypatch) -> None:
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:PBv4B:TEST123:4.4.1"),  # Called by PowerBoard.__init__
    ])
    mock_atexit = MockAtExit()
    monkeypatch.setattr('sr.robot3.power_board.atexit', mock_atexit)
    monkeypatch.setattr('sr.robot3.power_board.SerialWrapper', serial_wrapper)
    power_board = PowerBoard('test://')

    assert power_board._cleanup in mock_atexit._callbacks

    yield MockPowerBoard(serial_wrapper, power_board)

    # Test that we made all the expected calls
    assert serial_wrapper.request_index == len(serial_wrapper.responses)


def test_power_board_identify(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that we can create a power board with a mock serial wrapper.

    Uses the identify method to test that the mock serial wrapper is working.
    """
    serial_wrapper = powerboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:PBv4B:TEST456:4.4.1"),
    ])
    power_board = powerboard_serial.power_board

    # Test that the port was correctly passed to the mock serial wrapper init
    assert serial_wrapper._port == 'test://'

    # Test that the identity is correctly set from the first *IDN? response
    assert power_board._identity.board_type == "PBv4B"
    assert power_board._identity.asset_tag == "TEST123"

    # Test identify returns a fresh identity
    assert power_board.identify().asset_tag == "TEST456"


def test_power_board(powerboard_serial: MockPowerBoard) -> None:
    """
    Test the general power board methods.
    """
    serial_wrapper = powerboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:PBv4B:TEST456:4.4.1"),
        ("*IDN?", "Student Robotics:PBv4B:TEST456:4.4.1"),
        ("BATT:I?", "1234"),
        ("BATT:V?", "12450"),
        ("*STATUS?", "0,0,0,0,0,0,0:39:0:5234"),
        ("*STATUS?", "0,0,0,0,0,0,0:39:0:5234"),
        ("*STATUS?", "0,0,0,0,0,0,0:39:0:5234"),
        ("BTN:START:GET?", "0:1"),
        ("NOTE:1047:100", "ACK"),
        ("NOTE:261:234", "ACK"),
        ("*RESET", "ACK"),
    ])
    power_board = powerboard_serial.power_board

    # Test identify returns a fresh identity
    assert power_board.identify().asset_tag == "TEST456"

    # Test that we can get the power board version
    assert power_board.identify().sw_version == "4.4.1"

    # Test that we can get the power board total current
    assert power_board.battery_sensor.current == 1.234

    # Test that we can get the power board total voltage
    assert power_board.battery_sensor.voltage == 12.45

    # Test that we can get the power board temperature
    assert power_board.status.temperature == 39

    # Test that we can get the power board fan status
    assert power_board.status.fan_running is False

    # Test that we can get the power board regulator voltage
    assert power_board.status.regulator_voltage == 5.234

    # Test that we can get the power board start button status
    assert power_board._start_button() is True

    # Test that we can play a note using the Note class
    power_board.piezo.buzz(Note.C6, 0.1)
    # And using a float
    power_board.piezo.buzz(261.63, 0.2345)

    # Test that we can reset the power board
    power_board.reset()


def test_power_board_leds(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that the power board LEDs are correctly mapped and their methods work.
    """
    power_board = powerboard_serial.power_board
    powerboard_serial.serial_wrapper._add_responses([
        ("LED:RUN:SET:1", "ACK"),
        ("LED:RUN:SET:0", "ACK"),
        ("LED:RUN:SET:F", "ACK"),
        ("LED:ERR:SET:1", "ACK"),
        ("LED:ERR:SET:0", "ACK"),
        ("LED:ERR:SET:F", "ACK"),
    ])

    # Test that we can set the power board run LED
    power_board._run_led.on()
    power_board._run_led.off()
    power_board._run_led.flash()

    # Test that we can set the power board error LED
    power_board._error_led.on()
    power_board._error_led.off()
    power_board._error_led.flash()


def test_power_board_outputs(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that the power board outputs are correctly mapped and their methods work.
    """
    power_board = powerboard_serial.power_board

    # Test that we can enable and disable the power board outputs
    powerboard_serial.serial_wrapper._add_responses([
        ("OUT:0:SET:1", "ACK"),
        ("OUT:1:SET:1", "ACK"),
        ("OUT:2:SET:1", "ACK"),
        ("OUT:3:SET:1", "ACK"),
        ("OUT:5:SET:1", "ACK"),
        ("OUT:6:SET:1", "ACK"),
        ("OUT:0:SET:0", "ACK"),
        ("OUT:1:SET:0", "ACK"),
        ("OUT:2:SET:0", "ACK"),
        ("OUT:3:SET:0", "ACK"),
        ("OUT:5:SET:0", "ACK"),
        ("OUT:6:SET:0", "ACK"),
    ])
    power_board.outputs.power_on()
    power_board.outputs.power_off()

    # Test that we can enable and disable the power board outputs individually
    powerboard_serial.serial_wrapper._add_responses([
        ("OUT:0:SET:1", "ACK"),
        ("OUT:0:SET:0", "ACK"),
        ("OUT:1:SET:1", "ACK"),
        ("OUT:1:SET:0", "ACK"),
        ("OUT:2:SET:1", "ACK"),
        ("OUT:2:SET:0", "ACK"),
        ("OUT:3:SET:1", "ACK"),
        ("OUT:3:SET:0", "ACK"),
        ("OUT:5:SET:1", "ACK"),
        ("OUT:5:SET:0", "ACK"),
        ("OUT:6:SET:1", "ACK"),
        ("OUT:6:SET:0", "ACK"),
    ])
    power_board.outputs[PowerOutputPosition.H0].is_enabled = True
    power_board.outputs[PowerOutputPosition.H0].is_enabled = False
    power_board.outputs[PowerOutputPosition.H1].is_enabled = True
    power_board.outputs[PowerOutputPosition.H1].is_enabled = False
    power_board.outputs[PowerOutputPosition.L0].is_enabled = True
    power_board.outputs[PowerOutputPosition.L0].is_enabled = False
    power_board.outputs[PowerOutputPosition.L1].is_enabled = True
    power_board.outputs[PowerOutputPosition.L1].is_enabled = False
    power_board.outputs[PowerOutputPosition.L3].is_enabled = True
    power_board.outputs[PowerOutputPosition.L3].is_enabled = False
    power_board.outputs[PowerOutputPosition.FIVE_VOLT].is_enabled = True
    power_board.outputs[PowerOutputPosition.FIVE_VOLT].is_enabled = False

    # Test that we can't enable or disable the brain output
    with pytest.raises(RuntimeError, match=r"Brain output cannot be controlled.*"):
        power_board.outputs[PowerOutputPosition.L2].is_enabled = True

    # Test that we can detect whether the power board outputs are enabled
    powerboard_serial.serial_wrapper._add_responses([
        ("OUT:0:GET?", "1"),
        ("OUT:1:GET?", "1"),
        ("OUT:2:GET?", "1"),
        ("OUT:3:GET?", "1"),
        ("OUT:4:GET?", "1"),
        ("OUT:5:GET?", "1"),
        ("OUT:6:GET?", "1"),
    ])
    assert power_board.outputs[PowerOutputPosition.H0].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.H1].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.L0].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.L1].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.L2].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.L3].is_enabled is True
    assert power_board.outputs[PowerOutputPosition.FIVE_VOLT].is_enabled is True

    # Test that we can detect whether the power board outputs are overcurrent
    powerboard_serial.serial_wrapper._add_responses([
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
        ("*STATUS?", "0,1,0,1,0,1,0:39:0:5234"),
    ])
    assert power_board.outputs[PowerOutputPosition.H0].overcurrent is False
    assert power_board.outputs[PowerOutputPosition.H1].overcurrent is True
    assert power_board.outputs[PowerOutputPosition.L0].overcurrent is False
    assert power_board.outputs[PowerOutputPosition.L1].overcurrent is True
    assert power_board.outputs[PowerOutputPosition.L2].overcurrent is False
    assert power_board.outputs[PowerOutputPosition.L3].overcurrent is True
    assert power_board.outputs[PowerOutputPosition.FIVE_VOLT].overcurrent is False

    # Test that we can detect output current
    powerboard_serial.serial_wrapper._add_responses([
        ("OUT:0:I?", "1100"),
        ("OUT:1:I?", "1200"),
        ("OUT:2:I?", "1300"),
        ("OUT:3:I?", "1400"),
        ("OUT:4:I?", "1500"),
        ("OUT:5:I?", "1600"),
        ("OUT:6:I?", "1700"),
    ])
    assert power_board.outputs[PowerOutputPosition.H0].current == 1.1
    assert power_board.outputs[PowerOutputPosition.H1].current == 1.2
    assert power_board.outputs[PowerOutputPosition.L0].current == 1.3
    assert power_board.outputs[PowerOutputPosition.L1].current == 1.4
    assert power_board.outputs[PowerOutputPosition.L2].current == 1.5
    assert power_board.outputs[PowerOutputPosition.L3].current == 1.6
    assert power_board.outputs[PowerOutputPosition.FIVE_VOLT].current == 1.7


def test_power_board_cleanup(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that the power board cleanup method works.
    """
    power_board = powerboard_serial.power_board
    powerboard_serial.serial_wrapper._add_responses([
        ("*RESET", "ACK"),
    ])

    # Test that the cleanup method calls the reset method
    power_board._cleanup()


def test_invalid_properties(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that settng invalid properties raise an AttributeError.
    """
    power_board = powerboard_serial.power_board

    with pytest.raises(AttributeError):
        power_board.invalid_property = 1

    with pytest.raises(AttributeError):
        power_board.outputs.invalid_property = 1

    with pytest.raises(AttributeError):
        power_board.outputs[0].invalid_property = 1

    with pytest.raises(AttributeError):
        power_board.battery_sensor.invalid_property = 1

    with pytest.raises(AttributeError):
        power_board.piezo.invalid_property = 1


def test_power_board_bounds_check(powerboard_serial: MockPowerBoard) -> None:
    """
    Test that the power board bounds checking works.
    """
    power_board = powerboard_serial.power_board
    # Invalid input values should be caught before sending to the power board
    # so no serial messages are expected

    # Test that we can't enable or disable an invalid output number
    with pytest.raises(IndexError):
        power_board.outputs[10].is_enabled = True
    with pytest.raises(IndexError):
        power_board.outputs[7000].is_enabled = True

    # Test that we can't enable or disable an invalid output number type
    with pytest.raises(TypeError):
        power_board.outputs[None].is_enabled = True
    with pytest.raises(TypeError):
        power_board.outputs[1.5].is_enabled = True

    # Test that we can't buzz an invalid note frequency
    with pytest.raises(ValueError):
        power_board.piezo.buzz(-1, 0.1)
    with pytest.raises(ValueError):
        power_board.piezo.buzz(0, 0.1)
    with pytest.raises(ValueError):
        power_board.piezo.buzz(10001, 0.1)

    # Test that we can't buzz an invalid note frequency type
    with pytest.raises(TypeError):
        power_board.piezo.buzz(None, 0.1)
    with pytest.raises(TypeError):
        power_board.piezo.buzz({}, 0.1)

    # Test that we can't buzz an invalid note duration
    with pytest.raises(ValueError):
        power_board.piezo.buzz(1000, -1)
    with pytest.raises(ValueError):
        power_board.piezo.buzz(1000, 10**100)

    # Test that we can't buzz an invalid note duration type
    with pytest.raises(TypeError):
        power_board.piezo.buzz(1000, None)
    with pytest.raises(TypeError):
        power_board.piezo.buzz(1000, {})


def test_power_board_discovery(monkeypatch) -> None:
    """
    Test that discovery finds power boards from USB serial ports.

    Test that different USB pid/vid combinations are ignored.
    """
    class ListPortInfo(NamedTuple):
        """A mock serial port info."""
        device: str
        manufacturer: str
        product: str
        serial_number: str
        vid: int
        pid: int

    def mock_comports() -> list[ListPortInfo]:
        ports = [
            ListPortInfo(
                device='test://1',
                manufacturer='Student Robotics',
                product='PBv4B',
                serial_number='TEST123',
                vid=0x1BDA,
                pid=0x0010,
            ),
            ListPortInfo(  # A power board with a different pid/vid
                device='test://3',
                manufacturer='Other',
                product='PBv4B',
                serial_number='OTHER',
                vid=0x1234,
                pid=0x5678,
            ),
            ListPortInfo(  # An unrelated device
                device='test://5',
                manufacturer='Student Robotics',
                product='OTHER',
                serial_number='TESTABC',
                vid=0x1BDA,
                pid=0x0010,
            ),
        ]
        return ports

    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:PBv4B:TEST123:4.4.1"),  # USB discovered board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.4.1"),  # USB invalid board
        ("*IDN?", "Student Robotics:PBv4B:TEST456:4.4.1"),  # Manually added board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.4.1"),  # Manual invalid board
    ])
    # mock atexit so we don't end up registering the cleanup method
    monkeypatch.setattr('sr.robot3.power_board.atexit', MockAtExit())
    monkeypatch.setattr('sr.robot3.power_board.SerialWrapper', serial_wrapper)
    monkeypatch.setattr('sr.robot3.power_board.comports', mock_comports)

    power_boards = PowerBoard._get_supported_boards(manual_boards=['test://2', 'test://4'])
    assert len(power_boards) == 2
    assert {'TEST123', 'TEST456'} == set(power_boards.keys())


def test_power_board_invalid_identity(monkeypatch) -> None:
    """
    Test that we raise an error if the power board returns an different board type.
    """
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:TestBoard:TEST123:4.4.1"),  # Called by PowerBoard.__init__
    ])
    monkeypatch.setattr('sr.robot3.power_board.SerialWrapper', serial_wrapper)

    with pytest.raises(
        IncorrectBoardError,
        match=re.escape("Board returned type 'TestBoard', expected 'PBv4B'"),
    ):
        PowerBoard('test://')


@pytest.mark.hardware
def test_physical_power_board_discovery() -> None:
    """Test that we can discover physical power boards."""
    power_boards = PowerBoard._get_supported_boards()
    assert len(power_boards) == 1, "Did not find exactly one power board."
    power_board = singular(power_boards)
    identity = power_board.identify()
    assert identity.board_type == "PBv4B", "Power board is not the correct type."
    asset_tag = identity.asset_tag
    assert power_board == power_boards[asset_tag], \
        "Singular power board is not the same as the one in the list of power boards."



from sr.robot3.timeout import kill_after_delay
import time

kill_after_delay(2)

while True:
    try:
        time.sleep(10)
    finally:
        # Imagine we wanted to do something else here
        pass



from __future__ import annotations

import logging
from datetime import datetime

from ..exceptions import BoardDisconnectionError, IncorrectBoardError
from ..serial_wrapper import SerialWrapper
from ..utils import BoardIdentity, get_simulator_boards

logger = logging.getLogger(__name__)

BAUDRATE = 115200


class TimeServer:
    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'TimeServer'.
        """
        return 'TimeServer'

    def __init__(
        self,
        serial_port: str,
        initial_identity: BoardIdentity | None = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()
        self._serial = SerialWrapper(
            serial_port,
            BAUDRATE,
            identity=initial_identity,
            # Disable the timeout so sleep works properly
            timeout=None,
        )

        self._identity = self.identify()
        if self._identity.board_type != self.get_board_type():
            raise IncorrectBoardError(self._identity.board_type, self.get_board_type())
        self._serial.set_identity(self._identity)

    @classmethod
    def initialise(cls) -> 'TimeServer' | None:
        # The filter here is the name of the emulated board in the simulator
        boards = get_simulator_boards('TimeServer')

        if not boards:
            return None

        board_info = boards[0]

        # Create board identity from the info given
        initial_identity = BoardIdentity(
            manufacturer='sbot_simulator',
            board_type=board_info.type_str,
            asset_tag=board_info.serial_number,
        )

        try:
            board = cls(board_info.url, initial_identity)
        except BoardDisconnectionError:
            logger.warning(
                f"Simulator specified time server at port {board_info.url!r}, "
                "could not be identified. Ignoring this device")
            return None
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None

        return board

    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        :return: The identity of the board.
        """
        response = self._serial.query('*IDN?')
        return BoardIdentity(*response.split(':'))

    def get_time(self) -> float:
        time_str = self._serial.query('TIME?')
        return datetime.fromisoformat(time_str).timestamp()

    def sleep(self, duration: float) -> None:
        if duration < 0:
            raise ValueError("sleep length must be non-negative")

        self._serial.query(f'SLEEP:{int(duration * 1000)}')



from __future__ import annotations
from typing import Optional

import pytest

from sr.robot3.utils import BoardIdentity


def pytest_addoption(parser):
    parser.addoption(
        "--hardware",
        action="store_true",
        help="Run hardware-in-the-loop tests.",
    )


def pytest_runtest_setup(item):
    if item.config.getoption("--hardware"):
        # Run hardware tests
        if 'hardware' not in item.keywords:
            pytest.skip("skipping non-hardware test")
    else:
        # Run unit tests
        if 'hardware' in item.keywords:
            pytest.skip("test requires physical boards connected and --hardware")


class MockSerialWrapper:
    """
    A class that mocks the sr.robot3.serial_wrapper.SerialWrapper class.

    Takes a list of 2-tuples of request and response strings.
    Asserts that each request is sent in order and returns the matching response.

    Implements the same interface as sr.robot3.serial_wrapper.SerialWrapper.
    """

    def __init__(self, responses: list[tuple[str, str]]) -> None:
        """Initialize the mock with a list of responses."""
        self.responses = responses
        self.request_index = 0
        self.identity = BoardIdentity()

    def _add_responses(self, responses: list[tuple[str, str]]) -> None:
        """Add more responses to the end of the list."""
        self.responses.extend(responses)

    def __call__(
        self,
        port: str,
        baud: int,
        timeout: float = 0.5,
        identity: BoardIdentity = BoardIdentity(),
        delay_after_connect: float = 0,
    ) -> 'MockSerialWrapper':
        """This will replace the original init method during the test."""
        self._port = port
        self._baudrate = baud
        return self

    def query(self, request: Optional[str], endl: str = '\n') -> str:
        """
        Mocks sending a command and returning the response.

        Asserts that the request is the next one in the list of expected requests.
        """
        # Assert that we have not run out of responses
        # and that the request is the next one we expect
        assert self.request_index < len(self.responses), f"Unexpected request: {request}"
        assert request == self.responses[self.request_index][0]

        # Fetch the response and increment the request index
        response = self.responses[self.request_index][1]
        self.request_index += 1
        return response

    def write(self, request: str, endl: str = '\n') -> None:
        """Send a command without waiting for a response."""
        _ = self.query(request, endl)

    def set_identity(self, identity: BoardIdentity) -> None:
        """Set the identity of the board."""
        self.identity = identity


class MockAtExit:
    def __init__(self):
        self._callbacks = []

    def register(self, callback):
        self._callbacks.append(callback)

    def unregister(self, callback):
        try:
            self._callbacks.remove(callback)
        except ValueError:
            pass



"""An implementation of a camera board using the april_vision library."""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Union

from april_vision import CalibratedCamera, Frame, FrameSource
from april_vision import Marker as AprilMarker
from april_vision import (
    Processor, USBCamera, __version__, calibrations,
    find_cameras, generate_marker_size_mapping,
)
from april_vision.helpers import Base64Sender
from numpy.typing import NDArray

from .marker import Marker
from .utils import (
    IN_SIMULATOR, Board, BoardIdentity, BoardInfo, get_simulator_boards,
)

PathLike = Union[Path, str]
LOGGER = logging.getLogger(__name__)

robot_calibrations = calibrations.copy()
# Include calibrations from the package after calibrations from the user's directory
robot_calibrations.insert(1, str(Path(__file__).parent.resolve() / 'calibrations'))


class AprilCamera(Board):
    """
    Virtual Camera Board for detecting fiducial markers.

    Additionally, it will do pose estimation, along with some calibration
    in order to determine the spatial positon and orientation of the markers
    that it has detected.

    :param camera_source: The source of the camera frames.
    :param calibration: The intrinsic calibration of the camera.
    :param serial_num: The serial number of the camera.
    :param name: The name of the camera.
    :param vidpid: The VID:PID of the camera.
    """
    __slots__ = ('_serial_num', '_cam')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of this board.

        :return: The literal string 'Camera'.
        """
        return 'Camera'

    @classmethod
    def _discover(cls) -> Dict[str, 'AprilCamera']:
        """
        Discover the connected cameras that have calibration data available.

        The calibration data from the april_vision library is included when searching.
        To add additional calibration data, add the paths to the environment variable
        `OPENCV_CALIBRATIONS`, separated by a colon.

        :return: A dict of cameras, keyed by their name and index.
        """
        if IN_SIMULATOR:
            return {
                camera_info.serial_number: cls.from_webots_camera(camera_info)
                for camera_info in get_simulator_boards('CameraBoard')
            }

        return {
            (serial := f"{camera_data.name} - {camera_data.index}"):
            cls.from_id(camera_data.index, camera_data=camera_data, serial_num=serial)
            for camera_data in find_cameras(robot_calibrations)
        }

    def __init__(
        self, camera_source: FrameSource,
        calibration: tuple[float, float, float, float] | None,
        serial_num: str,
        name: str,
        vidpid: str = "",
    ) -> None:
        # The processor handles the detection and pose estimation
        self._cam = Processor(
            camera_source,
            calibration=calibration,
            name=name,
            vidpid=vidpid,
            mask_unknown_size_tags=True,
        )
        self._serial_num = serial_num

    @classmethod
    def from_webots_camera(cls, camera_info: BoardInfo) -> 'AprilCamera':
        """
        Create a camera from a webots camera.

        :param camera_info: The information about the virtual camera,
                            including the url to connect to.
        :return: The camera object.
        """
        from .simulator.camera import WebotsRemoteCameraSource

        camera_source = WebotsRemoteCameraSource(camera_info)
        return cls(
            camera_source,
            calibration=camera_source.calibration,
            serial_num=camera_info.serial_number,
            name=camera_info.serial_number,
        )

    @classmethod
    def from_id(
        cls,
        camera_id: int,
        camera_data: CalibratedCamera,
        serial_num: str,
    ) -> 'AprilCamera':
        """
        Create a camera from an ID.

        :param camera_id: The ID of the camera to create.
        :param camera_data: The calibration data for the camera.
        :param serial_num: The serial number of the camera.
        :return: The camera object.
        """
        # The camera source handles the connection between the camera and the processor
        camera_source = USBCamera.from_calibration_file(
            camera_id,
            calibration_file=camera_data.calibration,
            vidpid=camera_data.vidpid,
        )
        return cls(
            camera_source,
            calibration=camera_source.calibration,
            serial_num=serial_num,
            name=camera_data.name,
            vidpid=camera_data.vidpid,
        )

    def identify(self) -> BoardIdentity:
        """
        Get the identity of the camera.

        The asset tag of the board is the camera name and index.
        The version is the version of the april_vision library.

        :return: The identity of the board.
        """
        return BoardIdentity(
            manufacturer='april_vision',
            board_type='camera',
            asset_tag=self._serial_num,
            sw_version=__version__,
        )

    def close(self) -> None:
        """
        Close the camera.

        The camera will no longer work after this method is called.
        """
        self._cam.close()

    def see(
        self,
        *,
        frame: Optional[NDArray] = None,
        save: Optional[PathLike] = None,
    ) -> List[Marker]:
        """
        Capture an image and identify fiducial markers.

        :param frame: An image to detect markers in, instead of capturing a new one,
        :param save: If given, save the annotated frame to the path.
                     This is given a JPEG extension if none is provided.
        :returns: list of markers that the camera could see.
        """
        if frame is None:
            frame = self._cam.capture()

        markers = self._cam.see(frame=frame)

        if save:
            self._cam.save(save, frame=frame, detections=markers)
        return [Marker.from_april_vision_marker(marker) for marker in markers]

    def capture(self, *, save: Optional[PathLike] = None) -> NDArray:
        """
        Get the raw image data from the camera.

        :param save: If given, save the annotated frame to the path.
                     This is given a JPEG extension if none is provided.
        :returns: Camera pixel data
        """
        raw_frame = self._cam.capture()
        if save:
            self._cam.save(save, frame=raw_frame, annotated=False)
        return raw_frame

    def _set_marker_sizes(
        self,
        tag_sizes: Union[float, Dict[int, float]],
    ) -> None:
        """
        Set the size of tags that are used for pose estimation.

        If a dict is given for tag_sizes, only marker IDs that are keys of the
        dict will be detected.

        :param tag_sizes: The size of the tags to use for pose estimation given in meters.
        """
        self._cam.set_marker_sizes(tag_sizes)

    def _set_detection_hook(
        self,
        callback: Callable[[Frame, List[AprilMarker]], None],
    ) -> None:
        """
        Setup a callback to be run after each detection.

        The callback will be passed the frame and the list of markers that were detected.

        :param callback: The function to run after each detection.
        """
        self._cam.detection_hook = callback

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial_num}>"


def _setup_cameras(
    tag_sizes: Dict[Iterable[int], int],
    publish_func: Optional[Callable[[str, bytes], None]] = None,
) -> Dict[str, AprilCamera]:
    """
    Find all connected cameras with calibration and configure tag sizes.

    Optionally set a callback to send a base64 encode JPEG bytestream of each
    image detection is run on.

    :param tag_sizes: The size of the tags to use for pose estimation given in millimeters
    :param publish_func: Optionally, a function to call with the base64 encoded JPEG bytestream
    :return: A dict of cameras, keyed by their name and index.
    """
    # Unroll the tag ID iterables and convert the sizes to meters
    expanded_tag_sizes = generate_marker_size_mapping(tag_sizes)

    if publish_func:
        frame_sender = Base64Sender(publish_func)

    cameras = AprilCamera._discover()

    for camera in cameras.values():
        # Set the tag sizes in the camera
        camera._set_marker_sizes(expanded_tag_sizes)
        if publish_func:
            camera._set_detection_hook(frame_sender.annotated_frame_hook)

    return cameras



import atexit
import struct

import cv2
import numpy as np
from april_vision import FrameSource
from numpy.typing import NDArray
from serial import serial_for_url

from ..utils import BoardInfo

HEADER_SIZE = 5  # 1 byte for the type, 4 bytes for the length
IMAGE_TAG_ID = 0


class WebotsRemoteCameraSource(FrameSource):
    # Webots cameras include an alpha channel, this informs april_vision of how to handle it
    COLOURSPACE = cv2.COLOR_BGRA2GRAY

    def __init__(self, camera_info: BoardInfo) -> None:
        self.calibration = (0.0, 0.0, 0.0, 0.0)
        # Use pyserial to give a nicer interface for connecting to the camera socket
        self._serial = serial_for_url(camera_info.url, baudrate=115200, timeout=None)

        # Check the camera is connected
        response = self._make_request("*IDN?")
        if not response.split(b":")[1].lower().startswith(b"cam"):
            raise RuntimeError(f"Camera not connected to a camera, returned: {response!r}")

        # Get the calibration data for this camera
        response = self._make_request("CAM:CALIBRATION?")

        # The calibration data is returned as a string of floats separated by colons
        new_calibration = tuple(map(float, response.split(b":")))
        assert len(new_calibration) == 4, f"Invalid calibration data: {new_calibration}"
        self.calibration = new_calibration

        # Get the image size for this camera
        response = self._make_request("CAM:RESOLUTION?")
        self.image_size = tuple(map(int, response.split(b":")))
        assert len(self.image_size) == 2, f"Invalid image dimensions: {self.image_size}"

        atexit.register(self.close)

    def read(self, fresh: bool = True) -> NDArray:
        """
        The method for getting a new frame.

        :param fresh: Whether to flush the device's buffer before capturing
        the frame, unused.
        """
        self._serial.write(b"CAM:FRAME!\n")
        # The image is encoded as a TLV (Type, Length, Value) packet
        # so we need to read the header to get the type and length of the image
        header = self._serial.read(HEADER_SIZE)
        assert len(header) == HEADER_SIZE, f"Invalid header length: {len(header)}"
        img_tag, img_len = struct.unpack('>BI', header)
        assert img_tag == IMAGE_TAG_ID, f"Invalid image tag: {img_tag}"

        # Get the image data now we know the length
        img_data = self._serial.read(img_len)
        assert len(img_data) == img_len, f"Invalid image data length: {len(img_data)}"

        rgb_frame_raw: NDArray[np.uint8] = np.frombuffer(img_data, np.uint8)

        # Height is first, then width, then channels
        return rgb_frame_raw.reshape((self.image_size[1], self.image_size[0], 4)).copy()

    def close(self) -> None:
        """Close the underlying socket on exit."""
        self._serial.close()

    def _make_request(self, command: str) -> bytes:
        self._serial.write(command.encode() + b"\n")
        response = self._serial.readline()
        if not response.endswith(b"\n") or response.startswith(b"NACK:"):
            raise RuntimeError(f"Failed to communicate with camera, returned: {response!r}")
        return response



"""
SerialWrapper class for communicating with boards over serial.

This class is responsible for opening and closing the serial port,
and for handling port timeouts and disconnections.
"""
from __future__ import annotations

import atexit
import logging
import sys
import threading
import time
from functools import wraps
from typing import Callable, Optional, TypeVar, Union

import serial

from .exceptions import BoardDisconnectionError
from .logging import TRACE
from .utils import IN_SIMULATOR, BoardIdentity

logger = logging.getLogger(__name__)

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec
Param = ParamSpec("Param")
RetType = TypeVar("RetType")

E = TypeVar("E", bound=BaseException)
BASE_TIMEOUT: float | None

if IN_SIMULATOR:
    BASE_TIMEOUT = None  # Disable timeouts while in the simulator to allow for pausing
else:
    BASE_TIMEOUT = 0.5


def retry(
    times: int, exceptions: Union[type[E], tuple[type[E], ...]]
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """
    Decorator to retry a function a number of times on a given exception.
    If the function fails on the last attempt the exception is raised.

    This outer function is used to pass arguments to the decorator.

    :param times: The number of times to retry the function.
    :param exceptions: The exceptions to catch and retry on.
    :return: The templated decorator function.
    """
    def decorator(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
        """
        This is the actual decorator function that is returned by the decorator.

        The decorator retries the function a number of times on a given exception.
        If the function fails on the last attempt the exception is raised.

        :param func: The function to decorate.
        :return: The decorated function.
        """
        @wraps(func)
        def retryfn(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
            """
            The function that wraps the original function.

            This function retries the original function a number of times on a given exception.

            :return: The return value of the original function.
            """
            attempt = 0
            while attempt < times:
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    time.sleep(attempt * 0.5)
                    attempt += 1
            return func(*args, **kwargs)
        return retryfn
    return decorator


class SerialWrapper:
    def __init__(
        self,
        port: str,
        baud: int,
        timeout: float | None = BASE_TIMEOUT,
        identity: BoardIdentity = BoardIdentity(),
        delay_after_connect: float = 0,
    ):
        # Mutex serial port access to allow for multiple threads to use the same serial port
        self._lock = threading.Lock()
        self.identity = identity

        # Time to wait before sending data after connecting to a board
        self.delay_after_connect = delay_after_connect

        # pyserial serial port, the port will be opened on the first message
        self.serial = serial.serial_for_url(
            port,
            baudrate=baud,
            timeout=timeout,
            write_timeout=timeout,
            do_not_open=True,
        )

        atexit.register(self._disconnect, quiet=True)

    def start(self) -> None:
        """
        Helper method to open the serial port.

        This is not usually needed as the port will be opened on the first message.
        """
        self._connect()

    def stop(self) -> None:
        """
        Helper method to close the serial port.

        This is not usually needed as the port will be closed on garbage collection.
        """
        self._disconnect()

    @retry(times=3, exceptions=(BoardDisconnectionError, UnicodeDecodeError, OSError))
    def query(self, data: Optional[str], *, endl: str = '\n') -> str:
        """
        Send a command to the board and return the response.

        This method will automatically reconnect to the board and retry the command
        up to 3 times on serial errors.

        :param data: The data to write to the board.
        :param endl: The endline character to add to outgoing data and remove from
            received data
        :raises BoardDisconnectionError: If the serial connection fails during the transaction,
            including failing to respond to the command.
        :return: The response from the board with the trailing endline character removed.
        """
        with self._lock:
            if not self.serial.is_open:
                if not self._connect():
                    # If the serial port cannot be opened raise an error,
                    # this will be caught by the retry decorator
                    raise BoardDisconnectionError(
                        f'Connection to board {self.identity.board_type}:'
                        f'{self.identity.asset_tag} could not be established',
                    )

            try:
                if data is not None:
                    logger.log(TRACE, f'Serial write - {data!r}')
                    cmd = data + endl
                    self.serial.write(cmd.encode())

                response = self.serial.readline()
                try:
                    response_str = response.decode().rstrip('\n')
                except UnicodeDecodeError as e:
                    logger.warning(
                        f"Board {self.identity.board_type}:{self.identity.asset_tag} "
                        f"returned invalid characters: {response!r}")
                    raise e
                logger.log(
                    TRACE, f'Serial read  - {response_str!r}')

                if b'\n' not in response:
                    # If readline times out no error is raised, it returns an incomplete string
                    logger.warning((
                        f'Connection to board {self.identity.board_type}:'
                        f'{self.identity.asset_tag} timed out waiting for response'
                    ))
                    raise serial.SerialException('Timeout on readline')
            except serial.SerialException:
                # Serial connection failed, close the port and raise an error
                self._disconnect()
                raise BoardDisconnectionError((
                    f'Board {self.identity.board_type}:{self.identity.asset_tag} '
                    'disconnected during transaction'
                ))

            if response_str.startswith('NACK'):
                _, error_msg = response_str.split(':', maxsplit=1)
                logger.error(
                    f'Board {self.identity.board_type}:{self.identity.asset_tag} '
                    f'returned NACK on write command: {error_msg}'
                )
                raise RuntimeError(error_msg)

            return response_str

    def write(self, data: str, *, endl: str = '\n') -> None:
        """
        Send a command to the board that does not require a response.

        :param data: The data to write to the board.
        :param endl: The endline character to add to outgoing data and remove from
            received data
        :raises RuntimeError: If the board returns a NACK response,
            the firmware's error message is raised.
        """
        _ = self.query(data, endl=endl)

    def _connect(self) -> bool:
        """
        Connect to the class's serial port.

        This is called automatically when a message is sent to the board or the
        serial connection is lost.

        :raises serial.SerialException: If the serial port cannot be opened.
        :return: True if the serial port was opened successfully, False otherwise.
        """
        try:
            self.serial.open()
            if not IN_SIMULATOR:
                # Wait for the board to be ready to receive data
                # Certain boards will reset when the serial port is opened
                time.sleep(self.delay_after_connect)
            self.serial.reset_input_buffer()
        except serial.SerialException:
            logger.error(
                'Failed to connect to board '
                f'{self.identity.board_type}:{self.identity.asset_tag}'
            )
            return False

        logger.info(
            f'Connected to board {self.identity.board_type}: {self.identity.asset_tag}'
        )
        return True

    def _disconnect(self, quiet: bool = False) -> None:
        """
        Close the class's serial port.

        This is called automatically when the serial connection fails.
        The serial port will be reopened on the next message.
        """
        self.serial.close()
        if not quiet:
            logger.warning(
                f'Board {self.identity.board_type}:{self.identity.asset_tag} disconnected'
            )

    def set_identity(self, identity: BoardIdentity) -> None:
        """
        Stores the identity of the board this serial wrapper is connected to.

        This is used for logging purposes.

        :param identity: The identity of the board this serial wrapper is connected to.
        """
        self.identity = identity

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__qualname__} {self.serial.port!r} {self.identity.asset_tag!r}>"
        )



"""The Arduino module provides an interface to the Arduino firmware."""
from __future__ import annotations

import logging
from enum import Enum, IntEnum
from types import MappingProxyType
from typing import Optional

from serial.tools.list_ports import comports

from .exceptions import IncorrectBoardError
from .logging import log_to_debug
from .serial_wrapper import SerialWrapper
from .utils import (
    IN_SIMULATOR, Board, BoardIdentity,
    get_simulator_boards, get_USB_identity, map_to_float,
)

logger = logging.getLogger(__name__)
BAUDRATE = 115200
if IN_SIMULATOR:
    # Place each command on a new line in the simulator to simplify the implementation
    ENDLINE = '\n'
else:
    ENDLINE = ''

SUPPORTED_VID_PIDS = {
    (0x2341, 0x0043),  # Arduino Uno rev 3
    (0x2A03, 0x0043),  # Arduino Uno rev 3
    (0x1A86, 0x7523),  # Uno
    (0x10C4, 0xEA60),  # Ruggeduino
    (0x16D0, 0x0613),  # Ruggeduino
}


class GPIOPinMode(str, Enum):
    """The possible modes for a GPIO pin."""
    INPUT = 'INPUT'
    INPUT_PULLUP = 'INPUT_PULLUP'
    OUTPUT = 'OUTPUT'


class AnalogPins(IntEnum):
    """The analog pins on the Arduino."""
    A0 = 14
    A1 = 15
    A2 = 16
    A3 = 17
    A4 = 18
    A5 = 19


MODE_CHAR_MAP = {
    GPIOPinMode.INPUT: 'i',
    GPIOPinMode.INPUT_PULLUP: 'p',
    GPIOPinMode.OUTPUT: 'o',
}


DIGITAL_READ_MODES = {GPIOPinMode.INPUT, GPIOPinMode.INPUT_PULLUP, GPIOPinMode.OUTPUT}
DIGITAL_WRITE_MODES = {GPIOPinMode.OUTPUT}
ANALOG_READ_MODES = {GPIOPinMode.INPUT}


class Arduino(Board):
    """
    The Arduino board interface.

    This is intended to be used with Arduino Uno boards running the SR firmware.

    :param serial_port: The serial port to connect to.
    :param initial_identity: The identity of the board, as reported by the USB descriptor.
    """
    __slots__ = ('_serial_num', '_serial', '_pins', '_identity')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'Arduino'.
        """
        return 'Arduino'

    def __init__(
        self,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()

        # The arduino firmware cannot access the serial number reported in the USB descriptor
        self._serial_num = initial_identity.asset_tag
        self._serial = SerialWrapper(
            serial_port,
            BAUDRATE,
            identity=initial_identity,
            delay_after_connect=2,  # Wait for the board to reset after connecting
        )

        self._pins = (
            tuple(  # Pins 0 and 1 are reserved for serial comms
                Pin(self._serial, index, supports_analog=False, disabled=True)
                for index in range(2))
            + tuple(Pin(self._serial, index, supports_analog=False) for index in range(2, 14))
            + tuple(Pin(self._serial, index, supports_analog=True) for index in range(14, 20))
        )

        self._identity = self.identify()
        # Arduino board type is not validated to allow for custom firmwares
        if not self._identity.board_type.startswith('SR'):
            raise IncorrectBoardError(self._identity.board_type, 'SR*')
        self._serial.set_identity(self._identity)

    @classmethod
    def _get_valid_board(
        cls,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> Optional[Arduino]:
        """
        Attempt to connect to an Arduino and returning None if it fails identification.

        :param serial_port: The serial port to connect to.
        :param initial_identity: The identity of the board, as reported by the USB descriptor.

        :return: An Arduino object, or None if the board could not be identified.
        """
        try:
            board = cls(serial_port, initial_identity)
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None
        except Exception:
            if initial_identity is not None:
                if initial_identity.board_type == 'manual':
                    logger.warning(
                        f"Manually specified Arduino at port {serial_port!r} "
                        "could not be identified. Ignoring this device")
                elif initial_identity.manufacturer == 'sbot_simulator':
                    logger.warning(
                        f"Simulator specified arduino at port {serial_port!r} "
                        "could not be identified. Ignoring this device")
                return None

            logger.warning(
                f"Found Arduino-like serial port at {serial_port!r}, "
                "but it could not be identified. Ignoring this device")
            return None
        return board

    @classmethod
    def _get_simulator_boards(cls) -> MappingProxyType[str, Arduino]:
        """
        Get the simulator boards.

        :return: A mapping of board serial numbers to Arduinos
        """
        boards = {}
        # The filter here is the name of the emulated board in the simulator
        for board_info in get_simulator_boards('Arduino'):

            # Create board identity from the info given
            initial_identity = BoardIdentity(
                manufacturer='sbot_simulator',
                board_type=board_info.type_str,
                asset_tag=board_info.serial_number,
            )
            if (board := cls._get_valid_board(board_info.url, initial_identity)) is None:
                continue

            boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @classmethod
    def _get_supported_boards(
        cls,
        manual_boards: Optional[list[str]] = None,
        ignored_serials: Optional[list[str]] = None,
    ) -> MappingProxyType[str, Arduino]:
        """
        Discover the connected Arduinos, by matching the USB descriptor to SUPPORTED_VID_PIDS.

        :param manual_boards: A list of manually specified board port strings,
            defaults to None
        :param ignored_serials: A list of serial number to ignore during board discovery
        :return: A mapping of board serial numbers to Arduinos
        """
        if IN_SIMULATOR:
            return cls._get_simulator_boards()

        boards = {}
        if ignored_serials is None:
            ignored_serials = []
        serial_ports = comports()
        for port in serial_ports:
            if (port.vid, port.pid) in SUPPORTED_VID_PIDS:
                # Create board identity from USB port info
                initial_identity = get_USB_identity(port)
                if initial_identity.asset_tag in ignored_serials:
                    continue

                if (board := cls._get_valid_board(port.device, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board

        # Add any manually specified boards
        if isinstance(manual_boards, list):
            for manual_port in manual_boards:
                # Create board identity from the info given
                initial_identity = BoardIdentity(
                    board_type='manual',
                    asset_tag=manual_port,
                )

                if (board := cls._get_valid_board(manual_port, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @log_to_debug
    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        The asset tag of the board is the serial number from the USB descriptor.

        :return: The identity of the board.
        """
        response = self._serial.query('v', endl=ENDLINE)
        response_fields = response.split(':')

        # The arduino firmware cannot access the serial number reported in the USB descriptor
        return BoardIdentity(
            manufacturer=(
                "Student Robotics"
                if response_fields[0].startswith('SR')
                else "Arduino"),
            board_type=response_fields[0],
            asset_tag=self._serial_num,
            sw_version=response_fields[1],
        )

    @property
    @log_to_debug
    def pins(self) -> tuple[Pin, ...]:
        """
        The pins on the Arduino.

        :return: A tuple of the pins on the Arduino.
        """
        return self._pins

    @log_to_debug
    def command(self, command: str) -> str:
        """
        Send a command to the board.

        :param command: The command to send to the board.
        :return: The response from the board.
        """
        if IN_SIMULATOR:
            logger.warning("The command method is not fully supported in the simulator")
        return self._serial.query(command, endl=ENDLINE)

    def map_pin_number(self, pin_number: int) -> str:
        """
        Map the pin number to the the serial format.
        Pin numbers are sent as printable ASCII characters, with 0 being 'a'.

        :param pin_number: The pin number to encode.
        :return: The pin number in the serial format.
        :raises ValueError: If the pin number is invalid.
        """
        try:  # bounds check
            self.pins[pin_number]._check_if_disabled()
        except (IndexError, IOError):
            raise ValueError("Invalid pin provided") from None
        return chr(pin_number + ord('a'))

    @log_to_debug
    def ultrasound_measure(
        self,
        pulse_pin: int,
        echo_pin: int,
    ) -> int:
        """
        Measure the distance to an object using an ultrasound sensor.

        The sensor can only measure distances up to 4m.

        :param pulse_pin: The pin to send the ultrasound pulse from.
        :param echo_pin: The pin to read the ultrasound echo from.
        :raises ValueError: If either of the pins are invalid
        :return: The distance measured by the ultrasound sensor in mm.
        """
        try:  # bounds check
            pulse_id = self.map_pin_number(pulse_pin)
        except ValueError:
            raise ValueError("Invalid pulse pin provided") from None
        try:
            echo_id = self.map_pin_number(echo_pin)
        except ValueError:
            raise ValueError("Invalid echo pin provided") from None

        response = self._serial.query(f'u{pulse_id}{echo_id}', endl=ENDLINE)
        return int(response)

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Pin:
    """
    A pin on the Arduino.

    :param serial: The serial wrapper to use to communicate with the board.
    :param index: The index of the pin.
    :param supports_analog: Whether the pin supports analog reads.
    :param disabled: Whether the pin can be controlled.
    """
    __slots__ = ('_serial', '_index', '_supports_analog', '_disabled', '_mode')

    def __init__(
        self,
        serial: SerialWrapper,
        index: int,
        supports_analog: bool,
        disabled: bool = False
    ):
        self._serial = serial
        self._index = index
        self._supports_analog = supports_analog
        self._disabled = disabled
        self._mode = GPIOPinMode.INPUT

    @property
    @log_to_debug
    def mode(self) -> GPIOPinMode:
        """
        Get the mode of the pin.

        This returns the cached value since the board does not report this.

        :raises IOError: If this pin cannot be controlled.
        :return: The mode of the pin.
        """
        self._check_if_disabled()
        return self._mode

    @mode.setter
    @log_to_debug(setter=True)
    def mode(self, value: GPIOPinMode) -> None:
        """
        Set the mode of the pin.

        To do analog or digital reads set the mode to INPUT or INPUT_PULLUP.
        To do digital writes set the mode to OUTPUT.

        :param value: The mode to set the pin to.
        :raises IOError: If the pin mode is not a GPIOPinMode.
        :raises IOError: If this pin cannot be controlled.
        """
        self._check_if_disabled()
        if not isinstance(value, GPIOPinMode):
            raise IOError('Pin mode only supports being set to a GPIOPinMode')

        mode_char = MODE_CHAR_MAP.get(value)
        if mode_char is None:
            raise IOError(f'Pin mode {value} is not supported')
        self._serial.write(self._build_command(mode_char), endl=ENDLINE)
        self._mode = value

    @log_to_debug
    def digital_read(self) -> bool:
        """
        Perform a digital read on the pin.

        :raises IOError: If the pin's current mode does not support digital read
        :raises IOError: If this pin cannot be controlled.
        :return: The digital value of the pin.
        """
        self._check_if_disabled()
        if self.mode not in DIGITAL_READ_MODES:
            raise IOError(f'Digital read is not supported in {self.mode}')
        response = self._serial.query(self._build_command('r'), endl=ENDLINE)
        return response == 'h'

    @log_to_debug
    def digital_write(self, value: bool) -> None:
        """
        Write a digital value to the pin.

        :param value: The value to write to the pin.
        :raises IOError: If the pin's current mode does not support digital write.
        :raises IOError: If this pin cannot be controlled.
        """
        self._check_if_disabled()
        if self.mode not in DIGITAL_WRITE_MODES:
            raise IOError(f'Digital write is not supported in {self.mode}')
        if value:
            self._serial.write(self._build_command('h'), endl=ENDLINE)
        else:
            self._serial.write(self._build_command('l'), endl=ENDLINE)

    @log_to_debug
    def analog_read(self) -> float:
        """
        Get the analog voltage on the pin.

        This is returned in volts. Only pins A0-A5 support analog reads.

        :raises IOError: If the pin or its current mode does not support analog read.
        :raises IOError: If this pin cannot be controlled.
        :return: The analog voltage on the pin, ranges from 0 to 5.
        """
        ADC_MAX = 1023  # 10 bit ADC
        ADC_MIN = 0

        self._check_if_disabled()
        if self.mode not in ANALOG_READ_MODES:
            raise IOError(f'Analog read is not supported in {self.mode}')
        if not self._supports_analog:
            raise IOError('Pin does not support analog read')
        response = self._serial.query(self._build_command('a'), endl=ENDLINE)
        # map the response from the ADC range to the voltage range
        return map_to_float(int(response), ADC_MIN, ADC_MAX, 0.0, 5.0)

    def _check_if_disabled(self) -> None:
        """
        Check if the pin is disabled.

        :raises IOError: If the pin is disabled.
        """
        if self._disabled:
            raise IOError('This pin cannot be controlled.')

    def _map_pin_number(self) -> str:
        """
        Map the pin number to the the serial format.
        Pin numbers are sent as printable ASCII characters, with 0 being 'a'.

        :return: The pin number in the serial format.
        """
        return chr(self._index + ord('a'))

    def _build_command(self, cmd_char: str) -> str:
        """
        Generate the command to send to the board.

        :param cmd_char: The command character to send.
        :return: The command string.
        """
        return f'{cmd_char}{self._map_pin_number()}'

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__qualname__} "
            f"index={self._index} analog={self._supports_analog} "
            f"disabled={self._disabled} {self._serial}>"
        )


if __name__ == '__main__':  # pragma: no cover
    arduinos = Arduino._get_supported_boards()
    for serial_num, board in arduinos.items():
        print(serial_num)

        board.pins[4].mode = GPIOPinMode.INPUT
        board.pins[4].mode = GPIOPinMode.INPUT_PULLUP

        # Digital write
        board.pins[13].mode = GPIOPinMode.OUTPUT
        board.pins[13].digital_write(True)
        digital_write_value = board.pins[13].digital_read()
        print(f'Set pin 13 to output and set to {digital_write_value}')

        # Digital read
        board.pins[4].mode = GPIOPinMode.INPUT
        digital_read_value = board.pins[4].digital_read()
        print(f'Input 4 = {digital_read_value}')

        board.pins[5].mode = GPIOPinMode.INPUT_PULLUP
        digital_read_value = board.pins[5].digital_read()
        print(f'Input 5 = {digital_read_value}')

        # Analog read
        board.pins[AnalogPins.A0].mode = GPIOPinMode.INPUT
        analog_read_value = board.pins[AnalogPins.A0].analog_read()
        print(f'Analog input A0 = {analog_read_value}')



from __future__ import annotations
from typing import Iterable, Optional, Union

BCM: int = 11
BOARD: int = 10
BOTH: int = 33
FALLING: int = 32
HARD_PWM: int = 43
HIGH: int = 1
I2C: int = 42
IN: int = 1
LOW: int = 0
OUT: int = 0
PUD_DOWN: int = 21
PUD_OFF: int = 20
PUD_UP: int = 22
RPI_REVISION: int
SERIAL: int = 40
SPI: int = 41
UNKNOWN: int = -1
VERSION: str

def cleanup() -> None: ...
def setmode(mode: int) -> None: ...
def getmode() -> Optional[int]: ...
def gpio_function() -> None: ...
def input(pin: int) -> bool: ...
def output(pin: Union[int, Iterable[int]], state: Union[int, Iterable[int]]) -> None: ...
def setup(
    pin: Union[int, list[int]],
    direction: int,
    *,
    initial: Optional[int] = None,
    pull_up_down: Optional[int] = None,
) -> None: ...
def setwarnings(warnings: bool) -> None: ...

class PWM:
    def __init__(self, channel: int, frequency: int) -> None: ...
    def start(self, duty_cycle: int) -> None: ...
    def ChangeFrequency(self, frequency: int) -> None: ...
    def ChangeDutyCycle(self, duty_cycle: int) -> None: ...
    def stop(self) -> None: ...



"""The Arduino module provides an interface to the Arduino firmware."""
from __future__ import annotations

import logging
from contextlib import contextmanager
from time import sleep
from types import MappingProxyType
from typing import Generator, NamedTuple, Optional

from serial import Serial, SerialException, serial_for_url
from serial.tools.list_ports import comports

from .exceptions import BoardDisconnectionError
from .logging import log_to_debug
from .utils import Board, BoardIdentity, get_USB_identity

logger = logging.getLogger(__name__)
DEFAULT_BAUDRATE = 115200


class RawSerialDevice(NamedTuple):
    """A serial device to look for by serial number."""
    serial_number: str
    baudrate: int = DEFAULT_BAUDRATE


class RawSerial(Board):
    """
    The raw serial interface.

    This is intended to be used to communicate with Arduinos running custom
    firmwares and other custom boards.

    :param serial_port: The serial port to connect to.
    :param baudrate: The baudrate to use when connecting.
    :param identity: The identity of the board, as reported by the USB descriptor.
    """
    __slots__ = ('_serial', '_identity', 'delay_after_connect')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.
        For this class, it is unused.

        :return: The literal string 'RawSerial'.
        """
        return 'RawSerial'

    def __init__(
        self,
        serial_port: str,
        baudrate: int = DEFAULT_BAUDRATE,
        identity: Optional[BoardIdentity] = None,
    ) -> None:
        if identity is None:
            identity = BoardIdentity()

        # We know nothing about the board,
        # so we need to get the identity from the USB descriptor
        self._identity = identity

        # Open the serial port
        self._serial = serial_for_url(
            serial_port,
            baudrate=baudrate,
            timeout=0.5,  # 500ms timeout
            do_not_open=True,
        )

        # Certain boards will reset when the serial port is opened
        # Wait for the board to be ready to receive data
        self.delay_after_connect = 2
        self._reconnect()

    @classmethod
    def _get_supported_boards(
        cls,
        serial_devices: list[tuple[str, int]],
    ) -> MappingProxyType[str, RawSerial]:
        """
        Discover connected serial devices filtered by serial number.

        :param serial_devices: A list of serial ports to check (serial_number, baudrate),
            these are matched by serial number.
        :return: A mapping of board serial numbers to Arduinos
        """
        boards = {}
        serial_device_objs = [
            RawSerialDevice(serial_number, baudrate)
            for serial_number, baudrate in serial_devices
        ]
        device_lookup = {
            device.serial_number: device for device in serial_device_objs
        }

        serial_ports = comports()
        for port in serial_ports:
            found_device = device_lookup.get(port.serial_number)  # type: ignore[arg-type]
            if found_device is not None:
                # Create board identity from USB port info
                identity = get_USB_identity(port)
                identity = BoardIdentity(
                    manufacturer=identity.manufacturer,
                    board_type=identity.board_type,
                    asset_tag=identity.asset_tag,
                    sw_version=f"{port.vid:04x}:{port.pid:04x}",
                )

                try:
                    board = cls(port.device, found_device.baudrate, identity)
                except Exception as err:
                    logger.warning(
                        f"Failed to connect to {found_device.serial_number} because: {err}")
                    continue

                boards[identity.asset_tag] = board

        return MappingProxyType(boards)

    @log_to_debug
    def identify(self) -> BoardIdentity:
        """
        Get the cached identity of the port from the USB descriptor.

        :return: The identity of the port.
        """
        return self._identity

    @property
    @log_to_debug
    def port(self) -> Serial:
        """
        The pySerial Serial object.

        :return: The raw serial port.
        """
        return self._serial

    @log_to_debug
    def write(self, data: bytes) -> None:
        """
        Send bytes to the port.

        :param data: The bytes to send.
        """
        with self._handle_serial_reconnect():
            self._serial.write(data)

    @log_to_debug
    def read(self, size: int) -> bytes:
        """
        Read a line from the port.

        :param size: The number of bytes to read.
        :return: The line read from the port.
        """
        with self._handle_serial_reconnect():
            return self._serial.read(size)

    @log_to_debug
    def read_until(self, terminator: bytes = b'\n') -> bytes:
        """
        Read until a terminator is reached.

        :param terminator: The terminator character to stop reading at.
        :return: The data read from the port.
        """
        with self._handle_serial_reconnect():
            return self._serial.read_until(terminator)

    @contextmanager
    def _handle_serial_reconnect(self) -> Generator[None, None, None]:
        """
        A context manager to handle reconnecting the serial ports on faults.
        """
        if not self._serial.is_open:
            if not self._reconnect():
                # If the serial port cannot be opened raise an error
                raise BoardDisconnectionError(
                    f'Connection to port {self._identity.board_type}:'
                    f'{self._identity.asset_tag} could not be established',
                )
        try:
            yield
        except (SerialException, OSError):
            # Serial connection failed, close the port and raise an error
            self._serial.close()
            logger.warning(
                f'Port {self._identity.board_type}:{self._identity.asset_tag} disconnected',
            )
            raise BoardDisconnectionError(
                f'Port {self._identity.board_type}:{self._identity.asset_tag} '
                'disconnected during transaction',
            )

    def _reconnect(self) -> bool:
        """
        Connect to the class's serial port.

        This is called automatically when a message is sent to the board if the port is closed.

        :raises serial.SerialException: If the serial port cannot be opened.
        :return: True if the serial port was opened successfully, False otherwise.
        """
        try:
            self._serial.open()
            # Wait for the board to be ready to receive data
            # Certain boards will reset when the serial port is opened
            sleep(self.delay_after_connect)
            self._serial.reset_input_buffer()
        except SerialException:
            logger.error(
                'Failed to connect to port '
                f'{self._identity.board_type}:{self._identity.asset_tag}',
            )
            return False

        logger.info(
            f'Connected to port {self._identity.board_type}: {self._identity.asset_tag}',
        )
        return True

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"



"""
Test that the motor board can be created and used.

This test uses a mock serial wrapper to simulate the connection to the motor board.
"""
from __future__ import annotations

import re
from typing import NamedTuple

import pytest

from sr.robot3.exceptions import IncorrectBoardError
from sr.robot3.motor_board import MotorBoard, MotorPower
from sr.robot3.utils import singular

from .conftest import MockAtExit, MockSerialWrapper


class MockMotorBoard(NamedTuple):
    """A mock motor board."""

    serial_wrapper: MockSerialWrapper
    motor_board: MotorBoard


@pytest.fixture
def motorboard_serial(monkeypatch) -> None:
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:MCv4B:TEST123:4.4"),  # Called by MotorBoard.__init__
    ])
    mock_atexit = MockAtExit()
    monkeypatch.setattr('sr.robot3.motor_board.atexit', mock_atexit)
    monkeypatch.setattr('sr.robot3.motor_board.SerialWrapper', serial_wrapper)
    motor_board = MotorBoard('test://')

    assert motor_board._cleanup in mock_atexit._callbacks

    yield MockMotorBoard(serial_wrapper, motor_board)

    # Test that we made all the expected calls
    assert serial_wrapper.request_index == len(serial_wrapper.responses)


def test_motor_board_identify(motorboard_serial: MockMotorBoard) -> None:
    """
    Test that we can create a motor board with a mock serial wrapper.

    Uses the identify method to test that the mock serial wrapper is working.
    """
    serial_wrapper = motorboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:MCv4B:TEST456:4.4"),
    ])
    motor_board = motorboard_serial.motor_board

    # Test that the port was correctly passed to the mock serial wrapper init
    assert serial_wrapper._port == 'test://'

    # Test that the identity is correctly set from the first *IDN? response
    assert motor_board._identity.board_type == "MCv4B"
    assert motor_board._identity.asset_tag == "TEST123"

    # Test identify returns a fresh identity
    assert motor_board.identify().asset_tag == "TEST456"


def test_motor_board(motorboard_serial: MockMotorBoard) -> None:
    """
    Test the general functionality of the motor board.

    This test uses the mock serial wrapper to simulate the motor board.
    """
    motorboard = motorboard_serial.motor_board
    serial_wrapper = motorboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:MCv4B:TEST456:4.4"),
        ("*STATUS?", "0,1:5432"),
        ("*STATUS?", "0,1:5432"),
        ("*STATUS?", "0,1:5432"),
        ("*RESET", "ACK"),
    ])

    # Test that we can get the motor board version
    assert motorboard.identify().sw_version == "4.4"

    # Test that we can get the motor board status
    assert motorboard.status.input_voltage == 5.432
    assert motorboard.motors[0].in_fault is False
    assert motorboard.motors[1].in_fault is True

    # Test that we can reset the motor board
    motorboard.reset()


def test_motor_board_motors(motorboard_serial: MockMotorBoard) -> None:
    """
    Test the motor board motor functionality.
    """
    motorboard = motorboard_serial.motor_board
    serial_wrapper = motorboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("MOT:0:SET:500", "ACK"),
        ("MOT:1:SET:512", "ACK"),
        ("MOT:0:DISABLE", "ACK"),
        ("MOT:1:DISABLE", "ACK"),
        ("MOT:0:SET:0", "ACK"),
        ("MOT:1:SET:0", "ACK"),
        ("MOT:0:GET?", "1:123"),
        ("MOT:1:GET?", "1:456"),
        ("MOT:0:GET?", "0:100"),
        ("MOT:1:GET?", "0:0"),
        ("MOT:0:I?", "1234"),
        ("MOT:1:I?", "12345"),
    ])

    # Test that we can set the motor power
    motorboard.motors[0].power = 0.5
    motorboard.motors[1].power = 0.5123

    # Test that we can disable the motors
    motorboard.motors[0].power = MotorPower.COAST
    motorboard.motors[1].power = MotorPower.COAST

    # Test setting the power to BRAKE
    motorboard.motors[0].power = MotorPower.BRAKE
    motorboard.motors[1].power = MotorPower.BRAKE

    # Test that we can get the motor power
    assert motorboard.motors[0].power == 0.123
    assert motorboard.motors[1].power == 0.456

    # Test that we can get the motor enabled state
    assert motorboard.motors[0].power == MotorPower.COAST
    assert motorboard.motors[1].power == MotorPower.COAST

    # Test that we can get the motor current
    assert motorboard.motors[0].current == 1.234
    assert motorboard.motors[1].current == 12.345


def test_motor_board_bounds_check(motorboard_serial: MockMotorBoard) -> None:
    """
    Test that handling of out of bounds values is correct.
    """
    motorboard = motorboard_serial.motor_board
    # Invalid input values should be caught before sending to the motor board
    # so no serial messages are expected

    # Test that we handle invalid output numbers correctly
    with pytest.raises(IndexError):
        motorboard.motors[2].power = 0.5

    # Test that we handle invalid output number types correctly
    with pytest.raises(TypeError):
        motorboard.motors['a'].power = 0.5
    with pytest.raises(TypeError):
        motorboard.motors[0.5].power = 0.5
    with pytest.raises(TypeError):
        motorboard.motors[None].power = 0.5

    # Test that we handle invalid power values correctly
    with pytest.raises(ValueError):
        motorboard.motors[0].power = -1.2
    with pytest.raises(ValueError):
        motorboard.motors[0].power = 1.2
    with pytest.raises(ValueError):
        motorboard.motors[0].power = 100

    # Test that we handle invalid power value types correctly
    with pytest.raises(TypeError):
        motorboard.motors[0].power = 'a'
    with pytest.raises(TypeError):
        motorboard.motors[0].power = None
    with pytest.raises(TypeError):
        motorboard.motors[0].power = [0.5]


def test_motor_board_cleanup(motorboard_serial: MockMotorBoard) -> None:
    """
    Test that the motor board cleanup method works.
    """
    motor_board = motorboard_serial.motor_board
    motorboard_serial.serial_wrapper._add_responses([
        ("*RESET", "ACK"),
    ])

    # Test that the cleanup method calls the reset method
    motor_board._cleanup()


def test_invalid_properties(motorboard_serial: MockMotorBoard) -> None:
    """
    Test that settng invalid properties raise an AttributeError.
    """
    motor_board = motorboard_serial.motor_board

    with pytest.raises(AttributeError):
        motor_board.invalid_property = 1

    with pytest.raises(AttributeError):
        motor_board.motors.invalid_property = 1

    with pytest.raises(AttributeError):
        motor_board.motors[0].invalid_property = 1


def test_motor_board_discovery(monkeypatch) -> None:
    """
    Test that discovery finds motor boards from USB serial ports.

    Test that different USB pid/vid combinations are ignored.
    """
    class ListPortInfo(NamedTuple):
        """A mock serial port info."""
        device: str
        manufacturer: str
        product: str
        serial_number: str
        vid: int
        pid: int

    def mock_comports() -> list[ListPortInfo]:
        ports = [
            ListPortInfo(
                device='test://1',
                manufacturer='Student Robotics',
                product='MCv4B',
                serial_number='TEST123',
                vid=0x0403,
                pid=0x6001,
            ),
            ListPortInfo(  # A motor board with a different pid/vid
                device='test://3',
                manufacturer='Other',
                product='MCv4B',
                serial_number='OTHER',
                vid=0x1234,
                pid=0x5678,
            ),
            ListPortInfo(  # An unrelated device
                device='test://5',
                manufacturer='Student Robotics',
                product='OTHER',
                serial_number='TESTABC',
                vid=0x0403,
                pid=0x6001,
            ),
        ]
        return ports

    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:MCv4B:TEST123:4.4"),  # USB discovered board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.4"),  # USB invalid board
        ("*IDN?", "Student Robotics:MCv4B:TEST456:4.4"),  # Manually added board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.4"),  # Manual invalid board
    ])
    # mock atexit so we don't end up registering the cleanup method
    monkeypatch.setattr('sr.robot3.motor_board.atexit', MockAtExit())
    monkeypatch.setattr('sr.robot3.motor_board.SerialWrapper', serial_wrapper)
    monkeypatch.setattr('sr.robot3.motor_board.comports', mock_comports)

    motor_boards = MotorBoard._get_supported_boards(manual_boards=['test://2', 'test://4'])
    assert len(motor_boards) == 2
    assert {'TEST123', 'TEST456'} == set(motor_boards.keys())


def test_motor_board_invalid_identity(monkeypatch) -> None:
    """
    Test that we raise an error if the motor board returns an different board type.
    """
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:TestBoard:TEST123:4.4"),  # Called by MotorBoard.__init__
    ])
    monkeypatch.setattr('sr.robot3.motor_board.SerialWrapper', serial_wrapper)

    with pytest.raises(
        IncorrectBoardError,
        match=re.escape("Board returned type 'TestBoard', expected 'MCv4B'"),
    ):
        MotorBoard('test://')


@pytest.mark.hardware
def test_physical_motor_board_discovery() -> None:
    """Test that we can discover physical motor boards."""
    motor_boards = MotorBoard._get_supported_boards()
    assert len(motor_boards) == 1, "Did not find exactly one motor board."
    motor_board = singular(motor_boards)
    identity = motor_board.identify()
    assert identity.board_type == "MCv4B", "motor board is not the correct type."
    asset_tag = identity.asset_tag
    assert motor_board == motor_boards[asset_tag], \
        "Singular motor board is not the same as the one in the list of motor boards."



"""Exceptions used by the sr-robot3 package."""


class MetadataKeyError(KeyError):
    """
    Raised when trying to access a metadata key for which no value exists.

    :param key: The key that was not found
    """

    def __init__(self, key: str):
        self.key = key

    def __str__(self) -> str:
        return f"Key {self.key!r} not present in metadata"


class MetadataNotReadyError(RuntimeError):
    """Raised when trying to access metadata before wait_start has been called."""

    def __str__(self) -> str:
        return (
            "Metadata (zone, arena and mode) can only be used after wait_start has been called"
        )


class BoardDisconnectionError(IOError):
    """Raised when communication to a board fails and cannot be reestablished."""
    pass


class IncorrectBoardError(IOError):
    """
    Raised when a board returns the wrong board type in response to *IDN?.

    This is usually caused by a board being provided the wrong serial port.

    :param board_type: The board type returned by the board
    :param expected_type: The board type expected by the class
    """
    def __init__(self, board_type: str, expected_type: str):
        self.returned_type = board_type
        self.expected_type = expected_type

    def __str__(self) -> str:
        return f"Board returned type {self.returned_type!r}, expected {self.expected_type!r}"



"""The power board module provides an interface to the power board firmware over serial."""
from __future__ import annotations

import atexit
import logging
from enum import IntEnum
from time import sleep
from types import MappingProxyType
from typing import Callable, NamedTuple, Optional

from serial.tools.list_ports import comports

from .exceptions import IncorrectBoardError
from .logging import log_to_debug
from .serial_wrapper import SerialWrapper
from .utils import (
    IN_SIMULATOR, Board, BoardIdentity, float_bounds_check,
    get_simulator_boards, get_USB_identity,
)

logger = logging.getLogger(__name__)
BAUDRATE = 115200  # Since the power board is a USB device, this is ignored
_SLEEP_FN = sleep  # For use in the simulator


class PowerOutputPosition(IntEnum):
    """
    A mapping of output name to number of the PowerBoard outputs.

    The numbers here are the same as used in communication protocol with the PowerBoard.
    """
    H0 = 0
    H1 = 1
    L0 = 2
    L1 = 3
    L2 = 4
    L3 = 5
    FIVE_VOLT = 6


class PowerStatus(NamedTuple):
    """A named tuple containing the values of the power status output."""
    overcurrent: tuple[bool, ...]
    temperature: int
    fan_running: bool
    regulator_voltage: float
    other: list[str] = []

    @classmethod
    def from_status_response(cls, response: str) -> PowerStatus:
        """
        Create a PowerStatus object from the response to a status command.

        :param response: The response from a *STATUS? command.
        :raise TypeError: If the response is invalid.
        :return: A PowerStatus object.
        """
        oc_flags, temp, fan_running, raw_voltage, *other = response.split(':')
        return cls(
            overcurrent=tuple((x == '1') for x in oc_flags.split(',')),
            temperature=int(temp),
            fan_running=(fan_running == '1'),
            regulator_voltage=float(raw_voltage) / 1000,
            other=other,
        )


# This output is always on, and cannot be controlled via the API.
BRAIN_OUTPUT = PowerOutputPosition.L2


class PowerBoard(Board):
    """
    A class representing the power board interface.

    This class is intended to be used to communicate with the power board over serial.

    :param serial_port: The serial port to connect to.
    :param initial_identity: The identity of the board, as reported by the USB descriptor.
    """
    __slots__ = (
        '_serial', '_identity', '_outputs', '_battery_sensor',
        '_piezo', '_run_led', '_error_led')

    @staticmethod
    def get_board_type() -> str:
        """
        Return the type of the board.

        :return: The literal string 'PBv4B'.
        """
        return 'PBv4B'

    def __init__(
        self,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> None:
        if initial_identity is None:
            initial_identity = BoardIdentity()
        self._serial = SerialWrapper(serial_port, BAUDRATE, identity=initial_identity)

        self._outputs = Outputs(self._serial)
        self._battery_sensor = BatterySensor(self._serial)
        self._piezo = Piezo(self._serial)
        self._run_led = Led(self._serial, 'RUN')
        self._error_led = Led(self._serial, 'ERR')

        self._identity = self.identify()
        if self._identity.board_type != self.get_board_type():
            raise IncorrectBoardError(self._identity.board_type, self.get_board_type())
        self._serial.set_identity(self._identity)

        atexit.register(self._cleanup)

    @classmethod
    def _get_valid_board(
        cls,
        serial_port: str,
        initial_identity: Optional[BoardIdentity] = None,
    ) -> Optional[PowerBoard]:
        """
        Attempt to connect to a power board and returning None if it fails identification.

        :param serial_port: The serial port to connect to.
        :param initial_identity: The identity of the board, as reported by the USB descriptor.

        :return: A PowerBoard object, or None if the board could not be identified.
        """

        try:
            board = cls(serial_port, initial_identity)
        except IncorrectBoardError as err:
            logger.warning(
                f"Board returned type {err.returned_type!r}, "
                f"expected {err.expected_type!r}. Ignoring this device")
            return None
        except Exception:
            if initial_identity is not None:
                if initial_identity.board_type == 'manual':
                    logger.warning(
                        f"Manually specified power board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")
                elif initial_identity.manufacturer == 'sbot_simulator':
                    logger.warning(
                        f"Simulator specified power board at port {serial_port!r}, "
                        "could not be identified. Ignoring this device")

            logger.warning(
                f"Found power board-like serial port at {serial_port!r}, "
                "but it could not be identified. Ignoring this device")
            return None
        return board

    @classmethod
    def _get_simulator_boards(cls) -> MappingProxyType[str, PowerBoard]:
        """
        Get the simulator boards.

        :return: A mapping of board serial numbers to boards.
        """
        boards = {}
        # The filter here is the name of the emulated board in the simulator
        for board_info in get_simulator_boards('PowerBoard'):

            # Create board identity from the info given
            initial_identity = BoardIdentity(
                manufacturer='sbot_simulator',
                board_type=board_info.type_str,
                asset_tag=board_info.serial_number,
            )
            if (board := cls._get_valid_board(board_info.url, initial_identity)) is None:
                continue

            boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @classmethod
    def _get_supported_boards(
        cls,
        manual_boards: Optional[list[str]] = None,
        sleep_fn: Optional[Callable[[float], None]] = None,
    ) -> MappingProxyType[str, PowerBoard]:
        """
        Find all connected power boards.

        Ports are filtered to the USB vendor and product ID: 0x1BDA and 0x0010 respectively.

        :param manual_boards: A list of manually specified serial ports to also attempt
            to connect to, defaults to None
        :return: A mapping of serial numbers to power boards.
        """
        if sleep_fn is not None:
            global _SLEEP_FN
            _SLEEP_FN = sleep_fn
        if IN_SIMULATOR:
            return cls._get_simulator_boards()

        boards = {}
        serial_ports = comports()
        for port in serial_ports:
            # Filter to USB vendor and product ID of the SR v4 power board
            if port.vid == 0x1BDA and port.pid == 0x0010:
                # Create board identity from USB port info
                initial_identity = get_USB_identity(port)

                if (board := cls._get_valid_board(port.device, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board

        # Add any manually specified boards
        if isinstance(manual_boards, list):
            for manual_port in manual_boards:
                # Create board identity from the info given
                initial_identity = BoardIdentity(
                    board_type='manual',
                    asset_tag=manual_port,
                )

                if (board := cls._get_valid_board(manual_port, initial_identity)) is None:
                    continue

                boards[board._identity.asset_tag] = board
        return MappingProxyType(boards)

    @property
    def outputs(self) -> Outputs:
        """
        Return the outputs of the power board.

        :return: The outputs of the power board.
        """
        return self._outputs

    @property
    def battery_sensor(self) -> BatterySensor:
        """
        Return the battery sensor of the power board.

        The battery sensor is used to measure the voltage and total current draw
        from the battery.

        :return: The battery sensor of the power board.
        """
        return self._battery_sensor

    @property
    def piezo(self) -> Piezo:
        """
        Return the piezo of the power board.

        The piezo is used to produce audible tones.

        :return: The piezo of the power board.
        """
        return self._piezo

    @log_to_debug
    def identify(self) -> BoardIdentity:
        """
        Get the identity of the board.

        :return: The identity of the board.
        """
        response = self._serial.query('*IDN?')
        return BoardIdentity(*response.split(':'))

    @property
    @log_to_debug
    def status(self) -> PowerStatus:
        """
        Return the status of the power board.

        :return: The status of the power board.
        """
        response = self._serial.query('*STATUS?')
        return PowerStatus.from_status_response(response)

    @log_to_debug
    def reset(self) -> None:
        """
        Reset the power board.

        This turns off all outputs except the brain output and stops any running tones.
        """
        self._serial.write('*RESET')

    def _start_button(self) -> bool:
        """
        Return whether the start button has been pressed.

        This value latches until the button is read, so only shows that the
        button has been pressed since this method was last called.

        :return: Whether the start button has been pressed.
        """
        response: str = self._serial.query('BTN:START:GET?')
        internal, external = response.split(':')
        return (internal == '1') or (external == '1')

    def _cleanup(self) -> None:
        """
        Reset the power board and turn off all outputs when exiting.

        This method is registered as an exit handler and is called to ensure
        the power board is left in a safe state.
        """
        try:
            self.reset()
        except Exception:
            logger.warning("Failed to cleanup power board.")

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Outputs:
    """
    A class representing the outputs of the power board.

    The individual outputs are accessed using the index operator.
    e.g. ``outputs[PowerOutputPosition.H0]``
    This also contains helper methods for controlling all outputs at once.

    :param serial: The serial wrapper to use for communication.
    """
    __slots__ = ('_serial', '_outputs')

    def __init__(self, serial: SerialWrapper):
        self._serial = serial
        self._outputs = tuple(Output(serial, i) for i in range(7))

    def __getitem__(self, key: int) -> Output:
        return self._outputs[key]

    @log_to_debug
    def power_off(self) -> None:
        """
        Turn off all outputs except the brain output.

        This is also used to turn off the outputs when the board is reset.
        """
        for output in self._outputs:
            if output._index == BRAIN_OUTPUT:
                continue
            output.is_enabled = False

    @log_to_debug
    def power_on(self) -> None:
        """Turn on all outputs."""
        for output in self._outputs:
            if output._index == BRAIN_OUTPUT:
                continue
            output.is_enabled = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Output:
    """
    A class representing a single output of the power board.

    :param serial: The serial wrapper to use for communication.
    :param index: The index of the output to represent.
    """
    __slots__ = ('_serial', '_index')

    def __init__(self, serial: SerialWrapper, index: int):
        self._serial = serial
        self._index = index

    @property
    @log_to_debug
    def is_enabled(self) -> bool:
        """
        Return whether the output is enabled.

        Outputs are enabled at startup, but will be disabled if the output draws
        too much current.

        :return: Whether the output is enabled.
        """
        response = self._serial.query(f'OUT:{self._index}:GET?')
        return response == '1'

    @is_enabled.setter
    @log_to_debug(setter=True)
    def is_enabled(self, value: bool) -> None:
        """
        Set whether the output is enabled.

        Outputs that have been disabled due to overcurrent will not be enabled,
        but will not raise an error.

        :param value: Whether the output should be enabled.
        """
        if self._index == BRAIN_OUTPUT:
            # Changing the brain output will also raise a NACK from the firmware
            raise RuntimeError("Brain output cannot be controlled via this API.")
        if value:
            self._serial.write(f'OUT:{self._index}:SET:1')
        else:
            self._serial.write(f'OUT:{self._index}:SET:0')

    @property
    @log_to_debug
    def current(self) -> float:
        """
        Return the current draw of the output.

        This current measurement has a 10% tolerance.

        :return: The current draw of the output, in amps.
        """
        response = self._serial.query(f'OUT:{self._index}:I?')
        return float(response) / 1000

    @property
    @log_to_debug
    def overcurrent(self) -> bool:
        """
        Return whether the output is in an overcurrent state.

        This is set when the output draws more than its maximum current.
        Resetting the power board will clear this state.

        :return: Whether the output is in an overcurrent state.
        """
        response = self._serial.query('*STATUS?')
        return PowerStatus.from_status_response(response).overcurrent[self._index]

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} index={self._index} {self._serial}>"


class Led:
    """
    A class representing a single LED of the power board.

    :param serial: The serial wrapper to use for communication.
    :param led: The name of the LED to represent.
    """
    __slots__ = ('_serial', '_led')

    def __init__(self, serial: SerialWrapper, led: str):
        self._serial = serial
        self._led = led

    @log_to_debug
    def on(self) -> None:
        """
        Turn on the LED.
        """
        self._serial.write(f'LED:{self._led}:SET:1')

    @log_to_debug
    def off(self) -> None:
        """
        Turn off the LED.
        """
        self._serial.write(f'LED:{self._led}:SET:0')

    @log_to_debug
    def flash(self) -> None:
        """
        Set the LED to flash at 1Hz.
        """
        self._serial.write(f'LED:{self._led}:SET:F')

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__} led={self._led} {self._serial}>"


class BatterySensor:
    """
    A class representing the battery sensor of the power board.

    This is implemented using an INA219 current sensor on the power board.

    :param serial: The serial wrapper to use for communication.
    """
    __slots__ = ('_serial',)

    def __init__(self, serial: SerialWrapper):
        self._serial = serial

    @property
    @log_to_debug
    def voltage(self) -> float:
        """
        Return the voltage of the battery.

        :return: The voltage of the battery, in volts.
        """
        response = self._serial.query('BATT:V?')
        return float(response) / 1000

    @property
    @log_to_debug
    def current(self) -> float:
        """
        Return the current draw from the battery.

        :return: The current draw from the battery, in amps.
        """
        response = self._serial.query('BATT:I?')
        return float(response) / 1000

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Piezo:
    """
    A class representing the piezo of the power board.

    The piezo is used to produce audible tones.

    :param serial: The serial wrapper to use for communication.
    """
    __slots__ = ('_serial',)

    def __init__(self, serial: SerialWrapper):
        self._serial = serial

    @log_to_debug
    def buzz(self, frequency: float, duration: float, *, blocking: bool = False) -> None:
        """
        Produce a tone on the piezo.

        This method is non-blocking, and sending another tone while one is
        playing will cancel the first.

        :param frequency: The frequency of the tone, in Hz.
        :param duration: The duration of the tone, in seconds.
        :param blocking: Whether to block until the tone has finished playing,
        """
        frequency_int = int(float_bounds_check(
            frequency, 8, 10_000, "Frequency must be between 8 and 10000Hz"))
        duration_ms = int(float_bounds_check(
            duration * 1000, 0, 2**31 - 1,
            f"Duration is a float in seconds between 0 and {(2**31 - 1) / 1000:,.0f}"))

        cmd = f'NOTE:{frequency_int}:{duration_ms}'
        self._serial.write(cmd)

        if blocking:
            _SLEEP_FN(duration)

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self._serial}>"


class Note(IntEnum):
    """
    An enumeration of notes.

    An enumeration of notes from scientific pitch
    notation and their related frequencies in Hz.
    """

    C6 = 1047
    D6 = 1175
    E6 = 1319
    F6 = 1397
    G6 = 1568
    A6 = 1760
    B6 = 1976
    C7 = 2093
    D7 = 2349
    E7 = 2637
    F7 = 2794
    G7 = 3136
    A7 = 3520
    B7 = 3951
    C8 = 4186



import functools
import logging
import sys
from typing import Callable, Optional, TypeVar, Union, overload

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec
Param = ParamSpec("Param")
RetType = TypeVar("RetType")

E = TypeVar("E", bound=BaseException)

# Add a TRACE level to logging, below DEBUG
TRACE = 5


def add_trace_level() -> None:
    """Set logging to label messages at level 5 as "TRACE"."""
    logging.addLevelName(TRACE, "TRACE")


def setup_logging(debug_logging: bool, trace_logging: bool) -> None:
    """
    Set up logging for the program.

    The logging level is set to INFO by default, but can be set to DEBUG or TRACE.
    This set on the root logger, so all loggers will inherit this level.

    Logs are output to the console, with the format:
    <logger name> - <log level> - <message>

    :param debug_logging: Enable debug level logging output to the console.
    :param trace_logging: Enable trace level logging output to the console.
    """
    logformat = '%(name)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(fmt=logformat)
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    if not len(root_logger.handlers):
        # only add a handler if there were no handlers previously attached
        root_logger.addHandler(handler)

    if trace_logging:
        # Trace level is below debug, so debug logs will still be shown
        root_logger.setLevel(TRACE)
        logging.log(TRACE, "Trace Mode is enabled")
    elif debug_logging:
        root_logger.setLevel(logging.DEBUG)
        logging.debug("Debug Mode is enabled")
    else:
        root_logger.setLevel(logging.INFO)


@overload
def log_to_debug(
    func: None = None, *, setter: bool = False,
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    ...


@overload
def log_to_debug(func: Callable[Param, RetType]) -> Callable[Param, RetType]:
    ...


def log_to_debug(
    func: Optional[Callable[Param, RetType]] = None, *, setter: bool = False,
) -> Union[
    Callable[[Callable[Param, RetType]], Callable[Param, RetType]],
    Callable[Param, RetType],
]:
    """
    Wrap a function to log its arguments and return value at DEBUG level.

    Logging is to the function's module logger.

    :param func: A function to wrap in debug logging
    :param setter: Whether the function is a getter (i.e. returns no value)
    :return: The wrapped function
    """

    def arg_wrapper(_func: Callable[Param, RetType]) -> Callable[Param, RetType]:
        logger = logging.getLogger(_func.__module__)

        @functools.wraps(_func)
        def wrapper_debug(*args: Param.args, **kwargs: Param.kwargs) -> RetType:
            args_repr = [repr(a) for a in args]
            kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
            signature = ", ".join(args_repr + kwargs_repr)

            if setter:
                logger.debug(f"Calling setter {_func.__qualname__}({signature})")
            else:
                logger.debug(f"Calling {_func.__qualname__}({signature})")
            value = _func(*args, **kwargs)
            if not setter:
                logger.debug(f"{_func.__qualname__!r} returned {value!r}")

            return value

        return wrapper_debug

    if func is None:
        return arg_wrapper
    else:
        return arg_wrapper(func)



"""General utility functions and classes for the sr-robot3 package."""
from __future__ import annotations

import logging
import os
import signal
import socket
from abc import ABC, abstractmethod
from types import FrameType
from typing import Any, Mapping, NamedTuple, Optional, Type, TypeVar

from serial.tools.list_ports import comports
from serial.tools.list_ports_common import ListPortInfo

T = TypeVar('T')
logger = logging.getLogger(__name__)

IN_SIMULATOR = os.environ.get('WEBOTS_SIMULATOR', '') == '1'


class BoardIdentity(NamedTuple):
    """
    A container for the identity of a board.

    All the board firmwares should return this information in response to
    the *IDN? query.

    :param manufacturer: The manufacturer of the board
    :param board_type: The short name of the board, i.e. PBv4B
    :param asset_tag: The asset tag of the board,
        this should match what is printed on the board
    :param sw_version: The firmware version of the board
    """
    manufacturer: str = ""
    board_type: str = ""
    asset_tag: str = ""
    sw_version: str = ""


class BoardInfo(NamedTuple):
    """A container for the information about a board connection."""
    url: str
    serial_number: str
    type_str: str


class Board(ABC):
    """
    This is the base class for all boards.

    Slots are used to prevent adding attributes to the class at runtime.
    They must also be defined in subclasses.
    """
    __slots__ = ('_identity',)

    @staticmethod
    @abstractmethod
    def get_board_type() -> str:
        """
        The string that is expected to be returned by the board's firmware.

        This should match the value in the board_type field of the response to
        the *IDN? query. This is implemented as a static method because class-level
        attributes are not compatible with slots.

        :return: The board type name
        """
        pass  # pragma: no cover

    @abstractmethod
    def identify(self) -> BoardIdentity:
        """
        Return the board's identity as reported by the firmware.

        :return: The board's identity
        """
        pass  # pragma: no cover


class FinalMeta(type):
    """
    A meta class which ensures a given class cannot be subclassed.

    It's recommended to additionally use typing.final.
    """
    def __new__(cls, name: str, bases: tuple, classdict: dict) -> Type:
        if bases:
            raise TypeError(f"{name} cannot be sub-classed.")
        return super().__new__(cls, name, bases, dict(classdict))


def map_to_int(
        x: float,
        in_min: float,
        in_max: float,
        out_min: int,
        out_max: int,
) -> int:
    """
    Map a value from the input range to the output range, returning the value as an int.

    This is used to convert a float value to an integer value for sending to the board.

    :param x: The value to map
    :param in_min: The lower bound of the input range
    :param in_max: The upper bound of the input range
    :param out_min: The lower bound of the output range
    :param out_max: The upper bound of the output range
    :return: The mapped value
    """
    value = (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
    return int(value)


def map_to_float(
        x: int,
        in_min: int,
        in_max: int,
        out_min: float,
        out_max: float,
        precision: int = 3,
) -> float:
    """
    Map a value from the input range to the output range, returning the value as a float.

    This is used to convert an integer value from the board to a float value.

    :param x: The value to map
    :param in_min: The lower bound of the input range
    :param in_max: The upper bound of the input range
    :param out_min: The lower bound of the output range
    :param out_max: The upper bound of the output range
    :param precision: The number of decimal places to round the output to, defaults to 3
    :return: The mapped value
    """
    value = (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
    return round(value, precision)


def singular(container: Mapping[str, T]) -> T:
    """
    Extract the single item from a container of one item.

    This is used to access individual boards without needing their serial number.

    :param container: A mapping of connected boards of a type
    :raises RuntimeError: If there is not exactly one of this type of board connected
    :return: The connected board of the type
    """
    length = len(container)

    if length == 1:
        return list(container.values())[0]
    elif length == 0:
        raise RuntimeError('No boards of this type found')
    else:
        raise RuntimeError(f'expected only one to be connected, but found {length}')


def obtain_lock(lock_port: int = 10653) -> socket.socket:
    """
    Bind to a port to claim it and prevent another process using it.

    We use this to prevent multiple robot instances running on the same machine,
    as they would conflict with each other.

    :param lock_port: The port to bind to, defaults to 10653 to be compatible with sr-robot3
    :raises OSError: If the port failed to bind, indicating another robot instance is running
    :return: The bound socket
    """
    lock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # We bind to a port to claim it and prevent another process using it
    try:
        lock.bind(("localhost", lock_port))
    except OSError:
        raise OSError('Unable to obtain lock, Is another robot instance already running?')

    return lock


def float_bounds_check(value: Any, min_val: float, max_val: float, err_msg: str) -> float:
    """
    Test that a value can be converted to a float and is within the given bounds.

    Errors are caught and re-raised with a more descriptive message.

    :param value: The value to test
    :param min_val: The minimum allowed value
    :param max_val: The maximum allowed value
    :param err_msg: The error message to raise if the value is invalid or out of bounds
    :raises TypeError: If the value cannot be converted to a float
    :raises ValueError: If the value is out of bounds
    :return: The input value converted to a float
    """
    try:
        new_val = float(value)
    except ValueError as e:
        raise TypeError(err_msg) from e

    if (new_val < min_val) or (new_val > max_val):
        raise ValueError(err_msg)

    return new_val


def get_USB_identity(port: ListPortInfo) -> BoardIdentity:
    """
    Generate an approximate identity for a board using the USB descriptor.

    This data will be overridden by the firmware once communication is established,
    but is used for early logging messages and error handling.

    :param port: The USB port information from pyserial
    :return: An initial identity for the board
    """
    try:
        return BoardIdentity(
            manufacturer=port.manufacturer or "",
            board_type=port.product or "",
            asset_tag=port.serial_number or "",
        )
    except Exception:
        logger.warning(
            f"Failed to pull identifying information from serial device {port.device}")
        return BoardIdentity()


def ensure_atexit_on_term() -> None:
    """
    Ensure `atexit` triggers on `SIGTERM`.
    > The functions registered via [`atexit`] are not called when the program is
      killed by a signal not handled by Python
    """

    if signal.getsignal(signal.SIGTERM) != signal.SIG_DFL:
        # If a signal handler is already present for SIGTERM,
        # this is sufficient for `atexit` to trigger, so do nothing.
        return

    def handle_signal(handled_signum: int, frame: Optional[FrameType]) -> None:
        """
        Handle the given signal by outputting some text and terminating the process.
        This will trigger `atexit`.
        """
        logger.info(signal.strsignal(handled_signum))
        exit(1)

    # Add the null-ish signal handler
    signal.signal(signal.SIGTERM, handle_signal)


def get_simulator_boards(board_filter: str = '') -> list[BoardInfo]:
    """
    Get a list of all boards configured in the simulator.

    This is used to support discovery of boards in the simulator environment.

    :param board_filter: A filter to only return boards of a certain type
    :return: A list of board connection information
    """
    if 'WEBOTS_ROBOT' not in os.environ:
        return []

    simulator_data = os.environ['WEBOTS_ROBOT'].splitlines()
    simulator_boards = []

    for board_data in simulator_data:
        board_data = board_data.rstrip('/')
        board_fragment, serial_number = board_data.rsplit('/', 1)
        board_url, board_type = board_fragment.rsplit('/', 1)

        board_info = BoardInfo(url=board_url, serial_number=serial_number, type_str=board_type)

        if board_filter and board_info.type_str != board_filter:
            continue

        simulator_boards.append(board_info)

    return simulator_boards


def list_ports() -> None:
    """
    Print a list of all connected USB serial ports.

    Output is in the format:
    ```
    VID:PID - Manufacturer - Product - Serial Number
    ```
    """
    print("VID:PID - Manufacturer - Product - Serial Number")
    for serial_port in comports():
        if serial_port.vid and serial_port.pid:
            # Only print USB serial ports
            print(
                f"{serial_port.vid:04X}:{serial_port.pid:04X} - {serial_port.manufacturer} "
                f"- {serial_port.product} - {serial_port.serial_number}"
            )



"""
Test that the servo board can be created and used.

This test uses a mock serial wrapper to simulate the connection to the servo board.
"""
from __future__ import annotations

import re
from typing import NamedTuple

import pytest

from sr.robot3.exceptions import IncorrectBoardError
from sr.robot3.servo_board import ServoBoard
from sr.robot3.utils import singular

from .conftest import MockAtExit, MockSerialWrapper


class MockServoBoard(NamedTuple):
    """A mock servo board."""

    serial_wrapper: MockSerialWrapper
    servo_board: ServoBoard


@pytest.fixture
def servoboard_serial(monkeypatch) -> None:
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:SBv4B:TEST123:4.3"),  # Called by ServoBoard.__init__
    ])
    mock_atexit = MockAtExit()
    monkeypatch.setattr('sr.robot3.servo_board.atexit', mock_atexit)
    monkeypatch.setattr('sr.robot3.servo_board.SerialWrapper', serial_wrapper)
    servo_board = ServoBoard('test://')

    assert servo_board._cleanup in mock_atexit._callbacks

    yield MockServoBoard(serial_wrapper, servo_board)

    # Test that we made all the expected calls
    assert serial_wrapper.request_index == len(serial_wrapper.responses)


def test_servo_board_identify(servoboard_serial: MockServoBoard) -> None:
    """
    Test that we can create a servo board with a mock serial wrapper.

    Uses the identify method to test that the mock serial wrapper is working.
    """
    serial_wrapper = servoboard_serial.serial_wrapper
    serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:SBv4B:TEST456:4.3"),
    ])
    servo_board = servoboard_serial.servo_board

    # Test that the port was correctly passed to the mock serial wrapper init
    assert serial_wrapper._port == 'test://'

    # Test that the identity is correctly set from the first *IDN? response
    assert servo_board._identity.board_type == "SBv4B"
    assert servo_board._identity.asset_tag == "TEST123"

    # Test identify returns a fresh identity
    assert servo_board.identify().asset_tag == "TEST456"


def test_servo_board(servoboard_serial: MockServoBoard) -> None:
    """
    Test that the servo board functionality works.

    This test uses the mock serial wrapper to simulate the servo board.
    """
    servo_board = servoboard_serial.servo_board
    servoboard_serial.serial_wrapper._add_responses([
        ("*IDN?", "Student Robotics:SBv4B:TEST123:4.3"),
        ("*STATUS?", "1:0"),
        ("*RESET", "ACK"),
        ("SERVO:I?", "1234"),
        ("SERVO:V?", "5432"),
    ])

    # Test that we can get the servo board version
    assert servo_board.identify().sw_version == "4.3"

    # Test that we can get the servo board status
    assert servo_board.status() == (True, False)

    # Test that we can reset the servo board
    servo_board.reset()

    # Test that we can get the servo board servo current
    assert servo_board.current == 1.234

    # Test that we can get the servo board servo voltage
    assert servo_board.voltage == 5.432


def test_servo_board_servos(servoboard_serial: MockServoBoard) -> None:
    """
    Test that the servo board servo functionality works.
    """
    servo_board = servoboard_serial.servo_board
    servoboard_serial.serial_wrapper._add_responses([
        ("SERVO:0:SET:1050", "ACK"),
        ("SERVO:1:SET:1500", "ACK"),
        ("SERVO:0:SET:1000", "ACK"),
        ("SERVO:1:SET:1000", "ACK"),
        ("SERVO:0:SET:1100", "ACK"),
        ("SERVO:1:SET:2000", "ACK"),
        ("SERVO:0:GET?", "1025"),
        ("SERVO:1:GET?", "1750"),
        ("SERVO:0:DISABLE", "ACK"),
        ("SERVO:0:GET?", "0"),
        ("SERVO:1:DISABLE", "ACK"),
        ("SERVO:1:GET?", "0"),
    ])
    # Test that we can set the duty cycle limits
    servo_board.servos[0].set_duty_limits(1000, 1100)

    # Test that we can get the duty cycle limits
    assert servo_board.servos[0].get_duty_limits() == (1000, 1100)

    # Test that we can set the servo duty cycle
    servo_board.servos[0].position = 0
    servo_board.servos[1].position = 0

    servo_board.servos[0].position = -1.0
    servo_board.servos[1].position = -1.0
    servo_board.servos[0].position = 1.0
    servo_board.servos[1].position = 1.0

    # Test that we can get the servo duty cycle
    assert servo_board.servos[0].position == -0.5
    assert servo_board.servos[1].position == 0.5

    # Test that we can disable the servo
    servo_board.servos[0].disable()
    assert servo_board.servos[0].position is None
    servo_board.servos[1].position = None
    assert servo_board.servos[1].position is None


def test_servo_board_bounds_checking(servoboard_serial: MockServoBoard) -> None:
    """
    Test that handling of out of bounds values is correct.
    """
    servo_board = servoboard_serial.servo_board
    # Invalid input values should be caught before sending to the motor board
    # so no serial messages are expected

    # invalid servo number
    with pytest.raises(IndexError):
        servo_board.servos[20].position = 0

    # invalid servo number type
    with pytest.raises(TypeError):
        servo_board.servos['a'].position = 0
    with pytest.raises(TypeError):
        servo_board.servos[0.1].position = 0
    with pytest.raises(TypeError):
        servo_board.servos[None].position = 0

    # invalid servo value
    with pytest.raises(ValueError):
        servo_board.servos[0].position = -1.1
    with pytest.raises(ValueError):
        servo_board.servos[0].position = 1.1
    with pytest.raises(ValueError):
        servo_board.servos[0].position = -100

    # invalid servo value type
    with pytest.raises(TypeError):
        servo_board.servos[0].position = 'a'
    with pytest.raises(TypeError):
        servo_board.servos[0].position = {}

    # duty cycle limits
    with pytest.raises(ValueError):
        servo_board.servos[0].set_duty_limits(100, 1000)
    with pytest.raises(ValueError):
        servo_board.servos[0].set_duty_limits(1000, 10000)

    # duty cycle limits type
    with pytest.raises(TypeError):
        servo_board.servos[0].set_duty_limits('a', 1000)
    with pytest.raises(TypeError):
        servo_board.servos[0].set_duty_limits(1000, 'a')
    with pytest.raises(TypeError):
        servo_board.servos[0].set_duty_limits(1000, None)
    with pytest.raises(TypeError):
        servo_board.servos[0].set_duty_limits(None, 1000)
    with pytest.raises(TypeError):
        servo_board.servos[0].set_duty_limits(1000, 1000.1)


def test_servo_board_cleanup(servoboard_serial: MockServoBoard) -> None:
    """
    Test that the servo board cleanup method works.
    """
    servo_board = servoboard_serial.servo_board
    servoboard_serial.serial_wrapper._add_responses([
        ("*RESET", "ACK"),
    ])

    # Test that the cleanup method calls the reset method
    servo_board._cleanup()


def test_invalid_properties(servoboard_serial: MockServoBoard) -> None:
    """
    Test that settng invalid properties raise an AttributeError.
    """
    servo_board = servoboard_serial.servo_board

    with pytest.raises(AttributeError):
        servo_board.invalid_property = 1

    with pytest.raises(AttributeError):
        servo_board.servos.invalid_property = 1

    with pytest.raises(AttributeError):
        servo_board.servos[0].invalid_property = 1


def test_servo_board_discovery(monkeypatch) -> None:
    """
    Test that discovery finds servo boards from USB serial ports.

    Test that different USB pid/vid combinations are ignored.
    """
    class ListPortInfo(NamedTuple):
        """A mock serial port info."""
        device: str
        manufacturer: str
        product: str
        serial_number: str
        vid: int
        pid: int

    def mock_comports() -> list[ListPortInfo]:
        ports = [
            ListPortInfo(
                device='test://1',
                manufacturer='Student Robotics',
                product='SBv4B',
                serial_number='TEST123',
                vid=0x1BDA,
                pid=0x0011,
            ),
            ListPortInfo(  # A servo board with a different pid/vid
                device='test://3',
                manufacturer='Other',
                product='SBv4B',
                serial_number='OTHER',
                vid=0x1234,
                pid=0x5678,
            ),
            ListPortInfo(  # An unrelated device
                device='test://5',
                manufacturer='Student Robotics',
                product='OTHER',
                serial_number='TESTABC',
                vid=0x1BDA,
                pid=0x0011,
            ),
        ]
        return ports

    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:SBv4B:TEST123:4.3"),  # USB discovered board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.3"),  # USB invalid board
        ("*IDN?", "Student Robotics:SBv4B:TEST456:4.3"),  # Manually added board
        ("*IDN?", "Student Robotics:OTHER:TESTABC:4.3"),  # Manual invalid board
    ])
    # mock atexit so we don't end up registering the cleanup method
    monkeypatch.setattr('sr.robot3.servo_board.atexit', MockAtExit())
    monkeypatch.setattr('sr.robot3.servo_board.SerialWrapper', serial_wrapper)
    monkeypatch.setattr('sr.robot3.servo_board.comports', mock_comports)

    servo_boards = ServoBoard._get_supported_boards(manual_boards=['test://2', 'test://4'])
    assert len(servo_boards) == 2
    assert {'TEST123', 'TEST456'} == set(servo_boards.keys())


def test_servo_board_invalid_identity(monkeypatch) -> None:
    """
    Test that we raise an error if the servo board returns an different board type.
    """
    serial_wrapper = MockSerialWrapper([
        ("*IDN?", "Student Robotics:TestBoard:TEST123:4.3"),  # Called by ServoBoard.__init__
    ])
    monkeypatch.setattr('sr.robot3.servo_board.SerialWrapper', serial_wrapper)

    with pytest.raises(
        IncorrectBoardError,
        match=re.escape("Board returned type 'TestBoard', expected 'SBv4B'"),
    ):
        ServoBoard('test://')


@pytest.mark.hardware
def test_physical_servo_board_discovery() -> None:
    """Test that we can discover physical servo boards."""
    servo_boards = ServoBoard._get_supported_boards()
    assert len(servo_boards) == 1, "Did not find exactly one servo board."
    servo_board = singular(servo_boards)
    identity = servo_board.identify()
    assert identity.board_type == "SBv4B", "servo board is not the correct type."
    asset_tag = identity.asset_tag
    assert servo_board == servo_boards[asset_tag], \
        "Singular servo board is not the same as the one in the list of servo boards."
